# COPYRIGHT (C) 2020-2025 Nicotine+ Contributors
# COPYRIGHT (C) 2016-2017 Michael Labouebe <gfarmerfr@free.fr>
# COPYRIGHT (C) 2008-2010 quinox <quinox@users.sf.net>
# COPYRIGHT (C) 2006-2009 daelstorm <daelstorm@gmail.com>
# COPYRIGHT (C) 2003-2004 Hyriand <hyriand@thegraveyard.org>
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time

from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk

from pynicotine.config import config
from pynicotine.core import core
from pynicotine.events import events
from pynicotine.gtkgui.application import GTK_API_VERSION
from pynicotine.gtkgui.application import GTK_MINOR_VERSION
from pynicotine.gtkgui.widgets import clipboard
from pynicotine.gtkgui.widgets import ui
from pynicotine.gtkgui.widgets.combobox import ComboBox
from pynicotine.gtkgui.widgets.dialogs import EntryDialog
from pynicotine.gtkgui.widgets.filechooser import FileChooserSave
from pynicotine.gtkgui.widgets.iconnotebook import IconNotebook
from pynicotine.gtkgui.widgets.infobar import InfoBar
from pynicotine.gtkgui.widgets.popupmenu import PopupMenu
from pynicotine.gtkgui.widgets.popupmenu import UserPopupMenu
from pynicotine.gtkgui.widgets.textview import TextView
from pynicotine.gtkgui.widgets.theme import add_css_class
from pynicotine.gtkgui.widgets.theme import get_flag_icon_name
from pynicotine.gtkgui.widgets.theme import remove_css_class
from pynicotine.gtkgui.widgets.treeview import TreeView
from pynicotine.logfacility import log
from pynicotine.slskmessages import ConnectionType
from pynicotine.slskmessages import UserStatus
from pynicotine.utils import humanize
from pynicotine.utils import human_speed


class UserInfos(IconNotebook):

    def __init__(self, window):

        super().__init__(
            window,
            parent=window.userinfo_content,
            parent_page=window.userinfo_page
        )

        self.page = window.userinfo_page
        self.page.id = "userinfo"
        self.toolbar = window.userinfo_toolbar
        self.toolbar_start_content = window.userinfo_title
        self.toolbar_end_content = window.userinfo_end
        self.toolbar_default_widget = window.userinfo_entry

        self.userinfo_combobox = ComboBox(
            container=self.window.userinfo_title, has_entry=True, has_entry_completion=True,
            entry=self.window.userinfo_entry, item_selected_callback=self.on_show_user_profile
        )

        # Events
        for event_name, callback in (
            ("add-buddy", self.add_remove_buddy),
            ("ban-user", self.ban_unban_user),
            ("check-privileges", self.check_privileges),
            ("ignore-user", self.ignore_unignore_user),
            ("peer-connection-closed", self.peer_connection_error),
            ("peer-connection-error", self.peer_connection_error),
            ("quit", self.quit),
            ("remove-buddy", self.add_remove_buddy),
            ("server-disconnect", self.server_disconnect),
            ("unban-user", self.ban_unban_user),
            ("unignore-user", self.ignore_unignore_user),
            ("user-country", self.user_country),
            ("user-info-progress", self.user_info_progress),
            ("user-info-remove-user", self.remove_user),
            ("user-info-response", self.user_info_response),
            ("user-info-show-user", self.show_user),
            ("user-interests", self.user_interests),
            ("user-stats", self.user_stats),
            ("user-status", self.user_status)
        ):
            events.connect(event_name, callback)

    def quit(self):
        self.freeze()

    def destroy(self):
        self.userinfo_combobox.destroy()
        super().destroy()

    def on_focus(self, *_args):

        if self.window.current_page_id != self.window.userinfo_page.id:
            return True

        if self.get_n_pages():
            return True

        if self.window.userinfo_entry.is_sensitive():
            self.window.userinfo_entry.grab_focus()
            return True

        return False

    def on_remove_all_pages(self, *_args):
        core.userinfo.remove_all_users()

    def on_restore_removed_page(self, page_args):
        username, = page_args
        core.userinfo.show_user(username)

    def on_show_user_profile(self, *_args):

        username = self.window.userinfo_entry.get_text().strip()

        if not username:
            return

        self.window.userinfo_entry.set_text("")
        core.userinfo.show_user(username)

    def show_user(self, user, refresh=False, switch_page=True):

        page = self.pages.get(user)

        if page is None:
            self.pages[user] = page = UserInfo(self, user)

            self.append_page(page.container, user, focus_callback=page.on_focus,
                             close_callback=page.on_close, user=user)
            page.set_label(self.get_tab_label_inner(page.container))

        elif refresh:
            page.set_indeterminate_progress()

        if switch_page:
            self.set_current_page(page.container)
            self.window.change_main_page(self.window.userinfo_page)

    def remove_user(self, user):

        page = self.pages.get(user)

        if page is None:
            return

        page.clear()
        self.remove_page(page.container, page_args=(user,))
        del self.pages[user]
        page.destroy()

    def check_privileges(self, _msg):
        for page in self.pages.values():
            page.update_privileges_button_state()

    def ban_unban_user(self, user):

        page = self.pages.get(user)

        if page is not None:
            page.update_ban_button_state()

    def ignore_unignore_user(self, user):

        page = self.pages.get(user)

        if page is not None:
            page.update_ignore_button_state()

    def add_remove_buddy(self, user, *_args):

        page = self.pages.get(user)

        if page is not None:
            page.update_buddy_button_state()

    def peer_connection_error(self, username, conn_type, **_unused):

        page = self.pages.get(username)

        if page is None:
            return

        if conn_type == ConnectionType.PEER:
            page.peer_connection_error()

    def user_stats(self, msg):

        page = self.pages.get(msg.user)

        if page is not None:
            page.user_stats(msg)

    def user_status(self, msg):

        page = self.pages.get(msg.user)

        if page is not None:
            self.set_user_status(page.container, msg.user, msg.status)

    def user_country(self, user, country_code):

        page = self.pages.get(user)

        if page is not None:
            page.user_country(country_code)

    def user_interests(self, msg):

        page = self.pages.get(msg.user)

        if page is not None:
            page.user_interests(msg)

    def user_info_progress(self, user, _sock, position, total):

        page = self.pages.get(user)

        if page is not None:
            page.user_info_progress(position, total)

    def user_info_response(self, msg):

        page = self.pages.get(msg.username)

        if page is not None:
            page.user_info_response(msg)

    def server_disconnect(self, *_args):
        for user, page in self.pages.items():
            self.set_user_status(page.container, user, UserStatus.OFFLINE)


class UserInfo:

    def __init__(self, userinfos, user):

        (
            self.add_remove_buddy_label,
            self.ban_unban_user_button,
            self.ban_unban_user_label,
            self.container,
            self.country_button,
            self.country_icon,
            self.country_label,
            self.description_view_container,
            self.dislikes_list_container,
            self.edit_interests_button,
            self.edit_profile_button,
            self.free_upload_slots_label,
            self.gift_privileges_button,
            self.ignore_unignore_user_button,
            self.ignore_unignore_user_label,
            self.info_bar_container,
            self.interests_container,
            self.likes_list_container,
            self.picture_view,
            self.progress_bar,
            self.queued_uploads_label,
            self.refresh_button,
            self.retry_button,
            self.shared_files_label,
            self.shared_folders_label,
            self.upload_slots_label,
            self.upload_speed_label,
            self.user_info_container,
            self.user_label
        ) = ui.load(scope=self, path="userinfo.ui")

        self.userinfos = userinfos
        self.window = userinfos.window

        self.info_bar = InfoBar(parent=self.info_bar_container, button=self.retry_button)
        self.description_view = TextView(self.description_view_container, editable=False, vertical_margin=5)
        self.user_label.set_text(user)

        if GTK_API_VERSION >= 4:
            self.country_icon.set_pixel_size(21)
            self.picture = Gtk.Picture(can_shrink=True, focusable=True, hexpand=True, vexpand=True)
            self.picture_view.append(self.picture)  # pylint: disable=no-member

            if (GTK_API_VERSION, GTK_MINOR_VERSION) >= (4, 8):
                self.picture.set_content_fit(Gtk.ContentFit.CONTAIN)
            else:
                self.picture.set_keep_aspect_ratio(True)

        else:
            # Setting a pixel size of 21 results in a misaligned country flag
            self.country_icon.set_pixel_size(0)

            self.picture = Gtk.EventBox(can_focus=True, hexpand=True, vexpand=True, visible=True)
            self.picture.connect("draw", self.on_draw_picture)

            self.picture_view.add(self.picture)    # pylint: disable=no-member

        self.user = user
        self.picture_data = None
        self.picture_surface = None
        self.indeterminate_progress = False
        self.refreshing = False

        # Set up likes list
        self.likes_list_view = TreeView(
            self.window, parent=self.likes_list_container,
            activate_row_callback=self.on_likes_row_activated,
            columns={
                "likes": {
                    "column_type": "text",
                    "title": _("Likes"),
                    "default_sort_type": "ascending"
                }
            }
        )

        # Set up dislikes list
        self.dislikes_list_view = TreeView(
            self.window, parent=self.dislikes_list_container,
            activate_row_callback=self.on_dislikes_row_activated,
            columns={
                "dislikes": {
                    "column_type": "text",
                    "title": _("Dislikes"),
                    "default_sort_type": "ascending"
                }
            }
        )

        # Popup menus
        self.user_popup_menu = UserPopupMenu(
            self.window.application, callback=self.on_tab_popup, username=user, tab_name="userinfo"
        )
        self.user_popup_menu.add_items(
            ("", None),
            ("#" + _("Close All Tabs…"), self.on_close_all_tabs),
            ("#" + _("_Close Tab"), self.on_close)
        )

        def get_interest_items(list_view, column_id):
            return (
                ("$" + _("I _Like This"), self.window.interests.on_like_recommendation, list_view, column_id),
                ("$" + _("I _Dislike This"), self.window.interests.on_dislike_recommendation, list_view, column_id),
                ("", None),
                ("#" + _("_Recommendations for Item"), self.window.interests.on_recommend_item, list_view, column_id),
                ("#" + _("_Search for Item"), self.window.interests.on_recommend_search, list_view, column_id)
            )

        self.likes_popup_menu = PopupMenu(self.window.application, self.likes_list_view.widget,
                                          self.on_popup_likes_menu)
        self.likes_popup_menu.add_items(*get_interest_items(self.likes_list_view, "likes"))

        self.dislikes_popup_menu = PopupMenu(self.window.application, self.dislikes_list_view.widget,
                                             self.on_popup_dislikes_menu)
        self.dislikes_popup_menu.add_items(*get_interest_items(self.dislikes_list_view, "dislikes"))

        self.picture_popup_menu = PopupMenu(self.window.application, self.picture)
        self.picture_popup_menu.add_items(
            ("#" + _("Copy Picture"), self.on_copy_picture),
            ("#" + _("Save Picture"), self.on_save_picture),
            ("", None),
            ("#" + _("Remove"), self.on_hide_picture)
        )

        self.popup_menus = (
            self.user_popup_menu, self.likes_popup_menu, self.dislikes_popup_menu,
            self.picture_popup_menu
        )

        self.remove_picture()
        self.populate_stats()
        self.update_button_states()

    def clear(self):

        self.description_view.clear()
        self.likes_list_view.clear()
        self.dislikes_list_view.clear()
        self.remove_picture()

    def destroy(self):

        for menu in self.popup_menus:
            menu.destroy()

        self.info_bar.destroy()
        self.description_view.destroy()
        self.likes_list_view.destroy()
        self.dislikes_list_view.destroy()
        self.__dict__.clear()

        self.indeterminate_progress = False  # Stop progress bar timer

    def set_label(self, label):
        self.user_popup_menu.set_parent(label)

    # General #

    def populate_stats(self):

        country_code = core.users.countries.get(self.user)
        stats = core.users.watched.get(self.user)

        if stats is not None:
            speed = stats.upload_speed or 0
            files = stats.files
            folders = stats.folders
        else:
            speed = 0
            files = folders = None

        if speed > 0:
            self.upload_speed_label.set_text(human_speed(speed))

        if files is not None:
            self.shared_files_label.set_text(humanize(files))

        if folders is not None:
            self.shared_folders_label.set_text(humanize(folders))

        if country_code:
            self.user_country(country_code)

    def remove_picture(self):

        if GTK_API_VERSION >= 4:
            # Empty paintable to prevent container width from shrinking
            self.picture.set_paintable(Gdk.Paintable.new_empty(intrinsic_width=1, intrinsic_height=1))

        self.picture_data = None
        self.picture_surface = None

        self.hide_picture()

    def load_picture(self, data):

        if not data:
            self.remove_picture()
            return

        try:
            if GTK_API_VERSION >= 4:
                self.picture_data = Gdk.Texture.new_from_bytes(GLib.Bytes(data))
                self.picture.set_paintable(self.picture_data)
            else:
                data_stream = Gio.MemoryInputStream.new_from_bytes(GLib.Bytes(data))
                self.picture_data = GdkPixbuf.Pixbuf.new_from_stream(data_stream, cancellable=None)
                self.picture_surface = Gdk.cairo_surface_create_from_pixbuf(self.picture_data, scale=1, for_window=None)

        except Exception as error:
            log.add(_("Failed to load picture for user %(user)s: %(error)s"), {
                "user": self.user,
                "error": error
            })
            self.remove_picture()
            return

        self.show_picture()

    def show_picture(self):

        if GTK_API_VERSION == 3:
            self.user_info_container.set_hexpand(False)
            self.interests_container.set_hexpand(False)

        self.picture_view.set_visible(True)
        self.picture_view.set_hexpand(True)

        add_css_class(self.interests_container, "border-end")

    def hide_picture(self):

        if GTK_API_VERSION == 3:
            self.user_info_container.set_hexpand(True)
            self.interests_container.set_hexpand(True)

        self.picture_view.set_visible(False)
        self.picture_view.set_hexpand(False)

        remove_css_class(self.interests_container, "border-end")

    def peer_connection_error(self):

        if not self.refreshing:
            return

        self.info_bar.show_error_message(
            _("Unable to request information from user. Either you both have a closed listening "
              "port, the user is offline, or there's a temporary connectivity issue.")
        )
        self.set_finished()

    def pulse_progress(self, repeat=True):

        if not self.indeterminate_progress:
            return False

        self.progress_bar.pulse()
        return repeat

    def user_info_progress(self, position, total):

        if not self.refreshing:
            return

        self.indeterminate_progress = False

        if total <= 0 or position <= 0:
            fraction = 0.0

        elif position < total:
            fraction = float(position) / total

        else:
            fraction = 1.0

        self.progress_bar.set_fraction(fraction)

    def set_indeterminate_progress(self):

        if self.indeterminate_progress:
            return

        self.indeterminate_progress = self.refreshing = True

        self.progress_bar.get_parent().set_reveal_child(True)
        self.progress_bar.pulse()
        GLib.timeout_add(320, self.pulse_progress, False)
        GLib.timeout_add(1000, self.pulse_progress)

        self.info_bar.set_visible(False)
        self.refresh_button.set_sensitive(False)

        if core.users.login_status == UserStatus.OFFLINE and self.user != config.sections["server"]["login"]:
            self.peer_connection_error()

    def set_finished(self):

        self.indeterminate_progress = self.refreshing = False

        self.userinfos.request_tab_changed(self.container)
        self.progress_bar.set_fraction(1.0)
        self.progress_bar.get_parent().set_reveal_child(False)

        self.refresh_button.set_sensitive(True)

    # Button States #

    def update_local_buttons_state(self):

        local_username = core.users.login_username or config.sections["server"]["login"]

        for widget in (self.edit_interests_button, self.edit_profile_button):
            widget.set_visible(self.user == local_username)

        for widget in (self.ban_unban_user_button, self.ignore_unignore_user_button):
            widget.set_visible(self.user != local_username)

    def update_buddy_button_state(self):
        label = _("Remove _Buddy") if self.user in core.buddies.users else _("Add _Buddy")
        self.add_remove_buddy_label.set_text_with_mnemonic(label)

    def update_ban_button_state(self):
        label = _("Unban User") if core.network_filter.is_user_banned(self.user) else _("Ban User")
        self.ban_unban_user_label.set_text(label)

    def update_ignore_button_state(self):
        label = _("Unignore User") if core.network_filter.is_user_ignored(self.user) else _("Ignore User")
        self.ignore_unignore_user_label.set_text(label)

    def update_privileges_button_state(self):
        self.gift_privileges_button.set_sensitive(bool(core.users.privileges_left))

    def update_button_states(self):

        self.update_local_buttons_state()
        self.update_buddy_button_state()
        self.update_ban_button_state()
        self.update_ignore_button_state()
        self.update_privileges_button_state()

    # Network Messages #

    def user_info_response(self, msg):

        if not self.refreshing:
            return

        if msg is None:
            return

        if msg.descr is not None:
            self.description_view.clear()
            self.description_view.append_line(msg.descr)

        self.free_upload_slots_label.set_text(_("Yes") if msg.slotsavail else _("No"))
        self.upload_slots_label.set_text(humanize(msg.totalupl))
        self.queued_uploads_label.set_text(humanize(msg.queuesize))

        self.queued_uploads_label.get_parent().set_visible(bool(msg.queuesize))

        self.picture_data = None
        self.load_picture(msg.pic)

        self.info_bar.set_visible(False)
        self.set_finished()

    def user_stats(self, msg):

        speed = msg.avgspeed or 0
        num_files = msg.files or 0
        num_folders = msg.dirs or 0

        h_speed = human_speed(speed) if speed > 0 else _("Unknown")
        h_num_files = humanize(num_files)
        h_num_folders = humanize(num_folders)

        if self.upload_speed_label.get_text() != h_speed:
            self.upload_speed_label.set_text(h_speed)

        if self.shared_files_label.get_text() != h_num_files:
            self.shared_files_label.set_text(h_num_files)

        if self.shared_folders_label.get_text() != h_num_folders:
            self.shared_folders_label.set_text(h_num_folders)

    def user_country(self, country_code):

        if not country_code:
            return

        country_name = core.network_filter.COUNTRIES.get(country_code, _("Unknown"))

        self.country_label.set_text(country_name)
        self.country_button.set_tooltip_text(f"{country_name} ({country_code})")

        icon_name = get_flag_icon_name(country_code)
        icon_args = (Gtk.IconSize.BUTTON,) if GTK_API_VERSION == 3 else ()  # pylint: disable=no-member

        self.country_icon.set_from_icon_name(icon_name, *icon_args)
        self.country_button.set_visible(bool(icon_name))

    def user_interests(self, msg):

        self.likes_list_view.clear()
        self.likes_list_view.freeze()
        self.dislikes_list_view.clear()
        self.dislikes_list_view.freeze()

        for like in msg.likes:
            self.likes_list_view.add_row([like], select_row=False)

        for hate in msg.hates:
            self.dislikes_list_view.add_row([hate], select_row=False)

        self.likes_list_view.unfreeze()
        self.dislikes_list_view.unfreeze()

    # Callbacks #

    def on_show_progress_bar(self, progress_bar):
        """Enables indeterminate progress bar mode when tab is active."""

        if not self.indeterminate_progress and progress_bar.get_fraction() <= 0.0:
            self.set_indeterminate_progress()

    def on_hide_progress_bar(self, progress_bar):
        """Disables indeterminate progress bar mode when switching to another tab."""

        if self.indeterminate_progress:
            self.indeterminate_progress = False
            progress_bar.set_fraction(0.0)

    def on_draw_picture(self, area, context):
        """Draws a centered picture that fills the drawing area."""

        area_width = area.get_allocated_width()
        area_height = area.get_allocated_height()
        picture_width = self.picture_surface.get_width()
        picture_height = self.picture_surface.get_height()

        scale_factor = min(area_width / picture_width, area_height / picture_height)
        translate_x = (area_width - (picture_width * scale_factor)) / 2
        translate_y = (area_height - (picture_height * scale_factor)) / 2

        context.translate(translate_x, translate_y)
        context.scale(scale_factor, scale_factor)
        context.set_source_surface(self.picture_surface, 0, 0)
        context.paint()

    def on_tab_popup(self, *_args):
        self.user_popup_menu.toggle_user_items()

    def on_popup_likes_menu(self, menu, *_args):
        self.window.interests.toggle_menu_items(menu, self.likes_list_view, column_id="likes")

    def on_popup_dislikes_menu(self, menu, *_args):
        self.window.interests.toggle_menu_items(menu, self.dislikes_list_view, column_id="dislikes")

    def on_edit_profile(self, *_args):
        self.window.application.on_preferences(page_id="user-profile")

    def on_edit_interests(self, *_args):
        self.window.change_main_page(self.window.interests_page)

    def on_likes_row_activated(self, *_args):
        self.window.interests.show_item_recommendations(self.likes_list_view, column_id="likes")

    def on_dislikes_row_activated(self, *_args):
        self.window.interests.show_item_recommendations(self.dislikes_list_view, column_id="dislikes")

    def on_send_message(self, *_args):
        core.privatechat.show_user(self.user)

    def on_show_ip_address(self, *_args):
        core.users.request_ip_address(self.user, notify=True)

    def on_browse_user(self, *_args):
        core.userbrowse.browse_user(self.user)

    def on_add_remove_buddy(self, *_args):

        if self.user in core.buddies.users:
            core.buddies.remove_buddy(self.user)
            return

        core.buddies.add_buddy(self.user)

    def on_ban_unban_user(self, *_args):

        if core.network_filter.is_user_banned(self.user):
            core.network_filter.unban_user(self.user)
            return

        core.network_filter.ban_user(self.user)

    def on_ignore_unignore_user(self, *_args):

        if core.network_filter.is_user_ignored(self.user):
            core.network_filter.unignore_user(self.user)
            return

        core.network_filter.ignore_user(self.user)

    def on_give_privileges_response(self, dialog, _response_id, _data):

        days = dialog.get_entry_value()

        if not days:
            return

        try:
            days = int(days)

        except ValueError:
            self.on_give_privileges(error=_("Please enter number of days."))
            return

        core.users.request_give_privileges(self.user, days)

    def on_give_privileges(self, *_args, error=None):

        core.users.request_check_privileges()

        if core.users.privileges_left is None:
            days = _("Unknown")
        else:
            days = core.users.privileges_left // 60 // 60 // 24

        message = (_("Gift days of your Soulseek privileges to user %(user)s (%(days_left)s):") %
                   {"user": self.user, "days_left": _("%(days)s days left") % {"days": days}})

        if error:
            message += "\n\n" + error

        EntryDialog(
            parent=self.window,
            title=_("Gift Privileges"),
            message=message,
            action_button_label=_("_Give Privileges"),
            callback=self.on_give_privileges_response
        ).present()

    def on_copy_picture(self, *_args):

        if self.picture_data is None:
            return

        clipboard.copy_image(self.picture_data)

    def on_save_picture_response(self, selected, *_args):

        file_path = next(iter(selected), None)

        if not file_path:
            return

        if GTK_API_VERSION >= 4:
            picture_bytes = self.picture_data.save_to_png_bytes().get_data()
        else:
            _success, picture_bytes = self.picture_data.save_to_bufferv(
                type="png", option_keys=[], option_values=[])

        core.userinfo.save_user_picture(file_path, picture_bytes)

    def on_save_picture(self, *_args):

        if self.picture_data is None:
            return

        current_date_time = time.strftime("%Y-%m-%d_%H-%M-%S")

        FileChooserSave(
            parent=self.window,
            callback=self.on_save_picture_response,
            initial_folder=core.downloads.get_default_download_folder(),
            initial_file=f"{self.user}_{current_date_time}.png"
        ).present()

    def on_hide_picture(self, *_args):
        self.hide_picture()

    def on_refresh(self, *_args):
        core.userinfo.show_user(self.user, refresh=True)

    def on_focus(self, *_args):
        self.userinfos.grab_focus()
        return True

    def on_close(self, *_args):
        core.userinfo.remove_user(self.user)

    def on_close_all_tabs(self, *_args):
        self.userinfos.remove_all_pages()



# COPYRIGHT (C) 2020-2025 Nicotine+ Contributors
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

from gi.repository import GdkPixbuf
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk

import pynicotine
from pynicotine.config import config
from pynicotine.core import core
from pynicotine.gtkgui.application import GTK_API_VERSION
from pynicotine.gtkgui.application import GTK_GUI_FOLDER_PATH
from pynicotine.gtkgui.widgets.theme import ICON_THEME
from pynicotine.gtkgui.widgets.window import Window
from pynicotine.logfacility import log
from pynicotine.slskmessages import UserStatus
from pynicotine.utils import encode_path
from pynicotine.utils import truncate_string_byte


class ImplementationUnavailable(Exception):
    pass


class BaseImplementation:

    def __init__(self, application):

        self.application = application
        self.menu_items = {}
        self.menu_item_id = 1
        self.activate_callback = application.on_window_hide_unhide
        self.is_visible = True

        self._create_menu()

    def _create_item(self, text=None, callback=None, check=False):

        item = {"id": self.menu_item_id, "visible": True}

        if text is not None:
            item["text"] = text

        if callback is not None:
            item["callback"] = callback

        if check:
            item["toggled"] = False

        self.menu_items[self.menu_item_id] = item
        self.menu_item_id += 1

        return item

    def _set_item_text(self, item, text):
        item["text"] = text

    def _set_item_visible(self, item, visible):
        item["visible"] = visible

    def _set_item_toggled(self, item, toggled):
        item["toggled"] = toggled

    def _create_menu(self):

        self.show_hide_item = self._create_item("placeholder", self.application.on_window_hide_unhide)

        self._create_item()

        self.downloads_item = self._create_item(_("Downloads"), self.application.on_downloads)
        self.uploads_item = self._create_item(_("Uploads"), self.application.on_uploads)

        self._create_item()

        self._create_item(_("Private Chat"), self.application.on_private_chat)
        self._create_item(_("Chat Rooms"), self.application.on_chat_rooms)
        self._create_item(_("Searches"), self.application.on_searches)

        self._create_item()

        self.away_item = self._create_item(_("Away"), self.application.on_away, check=True)
        self.connect_item = self._create_item(_("_Connect"), self.application.on_connect)
        self.disconnect_item = self._create_item(_("_Disconnect"), self.application.on_disconnect)

        self._create_item()

        self._create_item(_("Preferences"), self.application.on_preferences)
        self._create_item(_("_Quit"), self.application.on_quit_request)

    def _update_window_visibility(self):

        if self.application.window is None:
            return

        label = _("Hide Nicotine+") if self.application.window.is_visible() else _("Show Nicotine+")

        if self.show_hide_item.get("text") != label:
            self._set_item_text(self.show_hide_item, label)

    def _update_user_status(self):

        should_update = False
        connect_visible = core.users.login_status == UserStatus.OFFLINE
        away_visible = not connect_visible
        away_toggled = core.users.login_status == UserStatus.AWAY

        if self.connect_item.get("visible") != connect_visible:
            self._set_item_visible(self.connect_item, connect_visible)
            should_update = True

        if self.disconnect_item.get("visible") == connect_visible:
            self._set_item_visible(self.disconnect_item, not connect_visible)
            should_update = True

        if self.away_item.get("visible") != away_visible:
            self._set_item_visible(self.away_item, away_visible)
            should_update = True

        if self.away_item.get("toggled") != away_toggled:
            self._set_item_toggled(self.away_item, away_toggled)
            should_update = True

        if should_update:
            self._update_icon()

    def _update_icon(self):

        # Check for highlights, and display highlight icon if there is a highlighted room or private chat
        if (self.application.window
                and (self.application.window.chatrooms.highlighted_rooms
                     or self.application.window.privatechat.highlighted_users)):
            icon_name = "msg"

        elif core.users.login_status == UserStatus.ONLINE:
            icon_name = "connect"

        elif core.users.login_status == UserStatus.AWAY:
            icon_name = "away"

        else:
            icon_name = "disconnect"

        icon_name = f"{pynicotine.__application_id__}-{icon_name}"
        self._set_icon_name(icon_name)

    def _set_icon_name(self, icon_name):
        # Implemented in subclasses
        pass

    def _update_icon_theme(self):
        # Implemented in subclasses
        pass

    def update(self):

        self.is_visible = config.sections["ui"]["trayicon"]

        self._update_icon_theme()
        self._update_icon()
        self._update_window_visibility()
        self._update_user_status()

    def set_download_status(self, status):
        self._set_item_text(self.downloads_item, status)

    def set_upload_status(self, status):
        self._set_item_text(self.uploads_item, status)

    def show_notification(self, title, message, action=None, action_target=None, high_priority=False):
        # Implemented in subclasses
        pass

    def unload(self, is_shutdown=True):
        # Implemented in subclasses
        pass


class StatusNotifierImplementation(BaseImplementation):

    class DBusProperty:

        def __init__(self, name, signature, value):

            self.name = name
            self.signature = signature
            self.value = value

    class DBusSignal:

        def __init__(self, name, args):
            self.name = name
            self.args = args

    class DBusMethod:

        def __init__(self, name, in_args, out_args, callback):

            self.name = name
            self.in_args = in_args
            self.out_args = out_args
            self.callback = callback

    class DBusService:

        def __init__(self, interface_name, object_path, bus_type):

            self._interface_name = interface_name
            self._object_path = object_path

            self._bus = Gio.bus_get_sync(bus_type)
            self._registration_id = None
            self.properties = {}
            self.signals = {}
            self.methods = {}

        def register(self):

            xml_output = f"<node name='/'><interface name='{self._interface_name}'>"

            for property_name, prop in self.properties.items():
                xml_output += f"<property name='{property_name}' type='{prop.signature}' access='read'/>"

            for method_name, method in self.methods.items():
                xml_output += f"<method name='{method_name}'>"

                for in_signature in method.in_args:
                    xml_output += f"<arg type='{in_signature}' direction='in'/>"
                for out_signature in method.out_args:
                    xml_output += f"<arg type='{out_signature}' direction='out'/>"

                xml_output += "</method>"

            for signal_name, signal in self.signals.items():
                xml_output += f"<signal name='{signal_name}'>"

                for signature in signal.args:
                    xml_output += f"<arg type='{signature}'/>"

                xml_output += "</signal>"

            xml_output += "</interface></node>"

            registration_id = self._bus.register_object(
                object_path=self._object_path,
                interface_info=Gio.DBusNodeInfo.new_for_xml(xml_output).interfaces[0],
                method_call_closure=self.on_method_call,
                get_property_closure=self.on_get_property,
                set_property_closure=None
            )

            if not registration_id:
                raise GLib.Error(f"Failed to register object with path {self._object_path}")

            self._registration_id = registration_id

        def unregister(self):

            if self._registration_id is None:
                return

            self._bus.unregister_object(self._registration_id)
            self._registration_id = None

        def add_property(self, name, signature, value):
            self.properties[name] = StatusNotifierImplementation.DBusProperty(name, signature, value)

        def add_signal(self, name, args):
            self.signals[name] = StatusNotifierImplementation.DBusSignal(name, args)

        def add_method(self, name, in_args, out_args, callback):
            self.methods[name] = StatusNotifierImplementation.DBusMethod(name, in_args, out_args, callback)

        def _emit_signal(self, name, parameters=None):

            self._bus.emit_signal(
                destination_bus_name=None,
                object_path=self._object_path,
                interface_name=self._interface_name,
                signal_name=name,
                parameters=parameters
            )

        def on_method_call(self, _connection, _sender, _path, _interface_name, method_name, parameters, invocation):

            method = self.methods[method_name]
            out_parameters = method.callback(*parameters.unpack())

            invocation.return_value(out_parameters)

        def on_get_property(self, _connection, _sender, _path, _interface_name, property_name):
            prop = self.properties[property_name]
            return GLib.Variant(prop.signature, prop.value)

    class DBusMenuService(DBusService):

        def __init__(self, menu_items):

            super().__init__(
                interface_name="com.canonical.dbusmenu",
                object_path="/StatusNotifierItem/menu",
                bus_type=Gio.BusType.SESSION
            )

            self._items = menu_items

            for method_name, in_args, out_args, callback in (
                ("GetGroupProperties", ("ai", "as"), ("a(ia{sv})",), self.on_get_group_properties),
                ("GetLayout", ("i", "i", "as"), ("u", "(ia{sv}av)"), self.on_get_layout),
                ("Event", ("i", "s", "v", "u"), (), self.on_event),
            ):
                self.add_method(method_name, in_args, out_args, callback)

            for signal_name, value in (
                ("LayoutUpdated", ("u", "i")),
            ):
                self.add_signal(signal_name, value)

        def update_item(self, item):

            item_id = item["id"]

            if item_id not in self._items:
                return

            self._items[item_id] = item

            self._emit_signal(
                name="ItemsPropertiesUpdated",
                parameters=GLib.Variant.new_tuple(
                    GLib.Variant.new_array(
                        children=[
                            GLib.Variant.new_tuple(
                                GLib.Variant.new_int32(item_id),
                                GLib.Variant.new_array(children=self._serialize_item(item))
                            )
                        ],
                    ),
                    GLib.Variant.new_array(child_type=GLib.VariantType("(ias)"))
                )
            )

        @staticmethod
        def _serialize_item(item):

            def dict_entry(key, value):
                return GLib.Variant.new_dict_entry(
                    GLib.Variant.new_string(key),
                    GLib.Variant.new_variant(value)
                )

            props = []

            if "text" in item:
                props.append(dict_entry("label", GLib.Variant.new_string(item["text"])))

                if item.get("toggled") is not None:
                    props.append(dict_entry("toggle-type", GLib.Variant.new_string("checkmark")))
                    props.append(dict_entry("toggle-state", GLib.Variant.new_int32(int(item["toggled"]))))
            else:
                props.append(dict_entry("type", GLib.Variant.new_string("separator")))

            props.append(dict_entry("visible", GLib.Variant.new_boolean(item["visible"])))
            return props

        def on_get_group_properties(self, ids, _properties):

            item_properties = []
            requested_ids = set(ids)

            for idx, item in self._items.items():
                # According to the spec, if no IDs are requested, we should send the entire menu
                if requested_ids and idx not in requested_ids:
                    continue

                item_properties.append(
                    GLib.Variant.new_tuple(
                        GLib.Variant.new_int32(idx),
                        GLib.Variant.new_array(children=self._serialize_item(item))
                    )
                )

            return GLib.Variant.new_tuple(
                GLib.Variant.new_array(children=item_properties)
            )

        def on_get_layout(self, _parent_id, _recursion_depth, _property_names):

            revision = 0
            serialized_items = []

            for idx, item in self._items.items():
                serialized_item = GLib.Variant.new_variant(
                    GLib.Variant.new_tuple(
                        GLib.Variant.new_int32(idx),
                        GLib.Variant.new_array(children=self._serialize_item(item)),
                        GLib.Variant.new_array(child_type=GLib.VariantType("v"))
                    )
                )
                serialized_items.append(serialized_item)

            return GLib.Variant.new_tuple(
                GLib.Variant.new_uint32(revision),
                GLib.Variant.new_tuple(
                    GLib.Variant.new_int32(0),
                    GLib.Variant.new_array(child_type=GLib.VariantType("{sv}")),
                    GLib.Variant.new_array(children=serialized_items)
                )
            )

        def on_event(self, idx, event_id, _data, _timestamp):
            if event_id == "clicked":
                self._items[idx]["callback"]()

    class StatusNotifierItemService(DBusService):

        def __init__(self, activate_callback, menu_items):

            super().__init__(
                interface_name="org.kde.StatusNotifierItem",
                object_path="/StatusNotifierItem",
                bus_type=Gio.BusType.SESSION
            )

            self.menu = StatusNotifierImplementation.DBusMenuService(menu_items)

            for property_name, signature, value in (
                ("Category", "s", "Communications"),
                ("Id", "s", pynicotine.__application_id__),
                ("Title", "s", pynicotine.__application_name__),
                ("ToolTip", "(sa(iiay)ss)", ("", [], pynicotine.__application_name__, "")),
                ("Menu", "o", "/StatusNotifierItem/menu"),
                ("ItemIsMenu", "b", False),
                ("IconName", "s", ""),
                ("IconThemePath", "s", ""),
                ("Status", "s", "Active")
            ):
                self.add_property(property_name, signature, value)

            for method_name, in_args, out_args, callback in (
                ("Activate", ("i", "i"), (), activate_callback),
                ("ProvideXdgActivationToken", ("s",), (), self.on_provide_activation_token)
            ):
                self.add_method(method_name, in_args, out_args, callback)

            for signal_name, value in (
                ("NewIcon", ()),
                ("NewIconThemePath", ("s",)),
                ("NewStatus", ("s",))
            ):
                self.add_signal(signal_name, value)

        def register(self):
            self.menu.register()
            super().register()

        def unregister(self):
            super().unregister()
            self.menu.unregister()

        def update_icon(self):
            self._emit_signal("NewIcon")

        def update_icon_theme(self, icon_path):

            self._emit_signal(
                name="NewIconThemePath",
                parameters=GLib.Variant.new_tuple(
                    GLib.Variant.new_string(icon_path)
                )
            )

        def update_status(self, status):

            self._emit_signal(
                name="NewStatus",
                parameters=GLib.Variant.new_tuple(
                    GLib.Variant.new_string(status)
                )
            )

        def on_provide_activation_token(self, token):
            Window.activation_token = token

    def __init__(self, application):

        super().__init__(application)

        self.tray_icon = None
        self.custom_icons = False

        try:
            self.bus = Gio.bus_get_sync(bus_type=Gio.BusType.SESSION)
            self.tray_icon = self.StatusNotifierItemService(
                activate_callback=self.activate_callback,
                menu_items=self.menu_items
            )
            self.tray_icon.register()

            self.bus.call_sync(
                bus_name="org.kde.StatusNotifierWatcher",
                object_path="/StatusNotifierWatcher",
                interface_name="org.kde.StatusNotifierWatcher",
                method_name="RegisterStatusNotifierItem",
                parameters=GLib.Variant.new_tuple(
                    GLib.Variant.new_string(self.bus.get_unique_name())
                ),
                reply_type=None,
                flags=Gio.DBusCallFlags.NONE,
                timeout_msec=1000,
                cancellable=None
            )

        except GLib.Error as error:
            self.unload()
            raise ImplementationUnavailable(f"StatusNotifier implementation not available: {error}") from error

    @staticmethod
    def _check_icon_path(icon_name, icon_path):
        """Check if tray icons exist in the specified icon path."""

        if not icon_path:
            return False

        icon_path_encoded = encode_path(icon_path)

        if not os.path.isdir(icon_path_encoded):
            return False

        icon_scheme = f"{pynicotine.__application_id__}-{icon_name}.".encode()

        try:
            with os.scandir(icon_path_encoded) as entries:
                for entry in entries:
                    if entry.is_file() and entry.name.startswith(icon_scheme):
                        return True

        except OSError as error:
            log.add_debug("Error accessing tray icon path %s: %s", (icon_path, error))

        return False

    def _get_icon_path(self):
        """Returns an icon path to use for tray icons, or None to fall back to
        system-wide icons."""

        self.custom_icons = False
        custom_icon_path = os.path.join(config.data_folder_path, ".nicotine-icon-theme")

        if hasattr(sys, "real_prefix") or sys.base_prefix != sys.prefix:
            # Virtual environment
            local_icon_path = os.path.join(sys.prefix, "share", "icons", "hicolor", "scalable", "apps")
        else:
            # Git folder
            local_icon_path = os.path.join(GTK_GUI_FOLDER_PATH, "icons", "hicolor", "scalable", "apps")

        for icon_name in ("away", "connect", "disconnect", "msg"):
            # Check if custom icons exist
            if config.sections["ui"]["icontheme"] and self._check_icon_path(icon_name, custom_icon_path):
                self.custom_icons = True
                return custom_icon_path

            # Check if local icons exist
            if self._check_icon_path(icon_name, local_icon_path):
                return local_icon_path

        return ""

    def _set_icon_name(self, icon_name):

        icon_name_property = self.tray_icon.properties["IconName"]

        if self.custom_icons:
            # Use alternative icon names to enforce custom icons, since system-wide icons take precedence
            icon_name = icon_name.replace(pynicotine.__application_id__, "nplus-tray")

        if icon_name_property.value != icon_name:
            icon_name_property.value = icon_name
            self.tray_icon.update_icon()

        if not self.is_visible:
            return

        status_property = self.tray_icon.properties["Status"]
        status = "Active"

        if status_property.value != status:
            status_property.value = status
            self.tray_icon.update_status(status)

    def _update_icon_theme(self):

        # If custom icon path was found, use it, otherwise we fall back to system icons
        icon_path = self._get_icon_path()
        icon_path_property = self.tray_icon.properties["IconThemePath"]

        if icon_path_property.value == icon_path:
            return

        icon_path_property.value = icon_path
        self.tray_icon.update_icon_theme(icon_path)

        if icon_path:
            log.add_debug("Using tray icon path %s", icon_path)

    def _set_item_text(self, item, text):
        super()._set_item_text(item, text)
        self.tray_icon.menu.update_item(item)

    def _set_item_visible(self, item, visible):
        super()._set_item_visible(item, visible)
        self.tray_icon.menu.update_item(item)

    def _set_item_toggled(self, item, toggled):
        super()._set_item_toggled(item, toggled)
        self.tray_icon.menu.update_item(item)

    def unload(self, is_shutdown=True):

        if self.tray_icon is None:
            return

        status_property = self.tray_icon.properties["Status"]
        status = "Passive"

        if status_property.value != status:
            status_property.value = status
            self.tray_icon.update_status(status)

        if is_shutdown:
            self.tray_icon.unregister()
            self.tray_icon = None


class Win32Implementation(BaseImplementation):
    """Windows NotifyIcon implementation.

    https://learn.microsoft.com/en-us/windows/win32/shell/notification-area
    https://learn.microsoft.com/en-us/windows/win32/shell/taskbar
    """

    WINDOW_CLASS_NAME = "NicotineTrayIcon"

    NIM_ADD = 0
    NIM_MODIFY = 1
    NIM_DELETE = 2

    NIF_MESSAGE = 1
    NIF_ICON = 2
    NIF_TIP = 4
    NIF_INFO = 16
    NIF_REALTIME = 64
    NIIF_NOSOUND = 16

    MIIM_STATE = 1
    MIIM_ID = 2
    MIIM_STRING = 64

    MFS_ENABLED = 0
    MFS_UNCHECKED = 0
    MFS_DISABLED = 3
    MFS_CHECKED = 8

    MFT_SEPARATOR = 2048

    WM_NULL = 0
    WM_DESTROY = 2
    WM_CLOSE = 16
    WM_COMMAND = 273
    WM_LBUTTONUP = 514
    WM_RBUTTONUP = 517
    WM_USER = 1024
    WM_TRAYICON = (WM_USER + 1)
    NIN_BALLOONHIDE = (WM_USER + 3)
    NIN_BALLOONTIMEOUT = (WM_USER + 4)
    NIN_BALLOONUSERCLICK = (WM_USER + 5)

    CS_VREDRAW = 1
    CS_HREDRAW = 2
    COLOR_WINDOW = 5
    IDC_ARROW = 32512

    WS_OVERLAPPED = 0
    WS_SYSMENU = 524288
    CW_USEDEFAULT = -2147483648

    IMAGE_ICON = 1
    LR_LOADFROMFILE = 16
    SM_CXSMICON = 49

    if sys.platform == "win32":
        from ctypes import Structure

        class WNDCLASSW(Structure):
            from ctypes import CFUNCTYPE, wintypes

            LPFN_WND_PROC = CFUNCTYPE(
                wintypes.INT,
                wintypes.HWND,
                wintypes.UINT,
                wintypes.WPARAM,
                wintypes.LPARAM
            )
            _fields_ = [
                ("style", wintypes.UINT),
                ("lpfn_wnd_proc", LPFN_WND_PROC),
                ("cb_cls_extra", wintypes.INT),
                ("cb_wnd_extra", wintypes.INT),
                ("h_instance", wintypes.HINSTANCE),
                ("h_icon", wintypes.HICON),
                ("h_cursor", wintypes.HANDLE),
                ("hbr_background", wintypes.HBRUSH),
                ("lpsz_menu_name", wintypes.LPCWSTR),
                ("lpsz_class_name", wintypes.LPCWSTR)
            ]

        class MENUITEMINFOW(Structure):
            from ctypes import wintypes

            _fields_ = [
                ("cb_size", wintypes.UINT),
                ("f_mask", wintypes.UINT),
                ("f_type", wintypes.UINT),
                ("f_state", wintypes.UINT),
                ("w_id", wintypes.UINT),
                ("h_sub_menu", wintypes.HMENU),
                ("hbmp_checked", wintypes.HBITMAP),
                ("hbmp_unchecked", wintypes.HBITMAP),
                ("dw_item_data", wintypes.LPVOID),
                ("dw_type_data", wintypes.LPWSTR),
                ("cch", wintypes.UINT),
                ("hbmp_item", wintypes.HBITMAP)
            ]

        class NOTIFYICONDATAW(Structure):
            from ctypes import wintypes

            _fields_ = [
                ("cb_size", wintypes.DWORD),
                ("h_wnd", wintypes.HWND),
                ("u_id", wintypes.UINT),
                ("u_flags", wintypes.UINT),
                ("u_callback_message", wintypes.UINT),
                ("h_icon", wintypes.HICON),
                ("sz_tip", wintypes.WCHAR * 128),
                ("dw_state", wintypes.DWORD),
                ("dw_state_mask", wintypes.DWORD),
                ("sz_info", wintypes.WCHAR * 256),
                ("u_version", wintypes.UINT),
                ("sz_info_title", wintypes.WCHAR * 64),
                ("dw_info_flags", wintypes.DWORD),
                ("guid_item", wintypes.CHAR * 16),
                ("h_balloon_icon", wintypes.HICON)
            ]

    def __init__(self, application):

        from ctypes import windll

        super().__init__(application)

        self._window_class = None
        self._h_wnd = None
        self._notify_id = None
        self._h_icon = None
        self._menu = None
        self._wm_taskbarcreated = windll.user32.RegisterWindowMessageW("TaskbarCreated")
        self._click_action = None
        self._click_action_target = None

        self._register_class()
        self._create_window()

    def _register_class(self):

        from ctypes import byref, windll

        self._window_class = self.WNDCLASSW(
            style=(self.CS_VREDRAW | self.CS_HREDRAW),
            lpfn_wnd_proc=self.WNDCLASSW.LPFN_WND_PROC(self.on_process_window_message),
            h_cursor=windll.user32.LoadCursorW(0, self.IDC_ARROW),
            hbr_background=self.COLOR_WINDOW,
            lpsz_class_name=self.WINDOW_CLASS_NAME
        )

        windll.user32.RegisterClassW(byref(self._window_class))

    def _unregister_class(self):

        if self._window_class is None:
            return

        from ctypes import windll

        windll.user32.UnregisterClassW(self.WINDOW_CLASS_NAME, self._window_class.h_instance)
        self._window_class = None

    def _create_window(self):

        from ctypes import windll

        style = self.WS_OVERLAPPED | self.WS_SYSMENU
        self._h_wnd = windll.user32.CreateWindowExW(
            0,
            self.WINDOW_CLASS_NAME,
            self.WINDOW_CLASS_NAME,
            style,
            0,
            0,
            self.CW_USEDEFAULT,
            self.CW_USEDEFAULT,
            0,
            0,
            0,
            None
        )

        windll.user32.UpdateWindow(self._h_wnd)

    def _destroy_window(self):

        if self._h_wnd is None:
            return

        from ctypes import windll

        windll.user32.DestroyWindow(self._h_wnd)
        self._h_wnd = None

    def _load_ico_buffer(self, icon_name, icon_size):

        ico_buffer = b""

        if GTK_API_VERSION >= 4:
            icon = ICON_THEME.lookup_icon(icon_name, fallbacks=None, size=icon_size, scale=1, direction=0, flags=0)
            icon_path = icon.get_file().get_path()

            if not icon_path:
                return ico_buffer

            pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(icon_path, icon_size, icon_size)
        else:
            icon = ICON_THEME.lookup_icon(icon_name, size=icon_size, flags=0)

            if not icon:
                return ico_buffer

            pixbuf = icon.load_icon()

        _success, ico_buffer = pixbuf.save_to_bufferv("ico")
        return ico_buffer

    def _load_h_icon(self, icon_name):

        from ctypes import windll

        # Attempt to load custom icons first
        icon_size = windll.user32.GetSystemMetrics(self.SM_CXSMICON)
        ico_buffer = self._load_ico_buffer(
            icon_name.replace(f"{pynicotine.__application_id__}-", "nplus-tray-"), icon_size)

        if not ico_buffer:
            # No custom icons present, fall back to default icons
            ico_buffer = self._load_ico_buffer(icon_name, icon_size)

        try:
            import tempfile
            file_handle = tempfile.NamedTemporaryFile(delete=False)

            with file_handle:
                file_handle.write(ico_buffer)

            return windll.user32.LoadImageA(
                0,
                encode_path(file_handle.name),
                self.IMAGE_ICON,
                icon_size,
                icon_size,
                self.LR_LOADFROMFILE
            )

        finally:
            os.remove(file_handle.name)

    def _destroy_h_icon(self):

        from ctypes import windll

        if self._h_icon:
            windll.user32.DestroyIcon(self._h_icon)
            self._h_icon = None

    def _update_notify_icon(self, title="", message="", icon_name=None, click_action=None,
                            click_action_target=None, high_priority=False):
        # pylint: disable=attribute-defined-outside-init,no-member

        if self._h_wnd is None:
            return

        if icon_name:
            self._destroy_h_icon()
            self._h_icon = self._load_h_icon(icon_name)

        if not self.is_visible and not (title or message):
            # When disabled by user, temporarily show tray icon when displaying a notification
            return

        from ctypes import byref, sizeof, windll

        notify_action = self.NIM_MODIFY

        if self._notify_id is None:
            self._notify_id = self.NOTIFYICONDATAW(
                cb_size=sizeof(self.NOTIFYICONDATAW),
                h_wnd=self._h_wnd,
                u_id=0,
                u_flags=(self.NIF_ICON | self.NIF_MESSAGE | self.NIF_TIP | self.NIF_INFO),
                u_callback_message=self.WM_TRAYICON,
                sz_tip=truncate_string_byte(pynicotine.__application_name__, byte_limit=127)
            )
            notify_action = self.NIM_ADD

        if config.sections["notifications"]["notification_popup_sound"]:
            self._notify_id.dw_info_flags &= ~self.NIIF_NOSOUND
        else:
            self._notify_id.dw_info_flags |= self.NIIF_NOSOUND

        if high_priority:
            self._notify_id.u_flags &= ~self.NIF_REALTIME
        else:
            self._notify_id.u_flags |= self.NIF_REALTIME

        self._notify_id.h_icon = self._h_icon
        self._notify_id.sz_info_title = truncate_string_byte(title, byte_limit=63, ellipsize=True)
        self._notify_id.sz_info = truncate_string_byte(message, byte_limit=255, ellipsize=True)

        self._click_action = click_action.replace("app.", "") if click_action else None
        self._click_action_target = GLib.Variant.new_string(click_action_target) if click_action_target else None

        windll.shell32.Shell_NotifyIconW(notify_action, byref(self._notify_id))

    def _remove_notify_icon(self):

        from ctypes import byref, windll

        if self._notify_id:
            windll.shell32.Shell_NotifyIconW(self.NIM_DELETE, byref(self._notify_id))
            self._notify_id = None

        self._destroy_menu()

    def _serialize_menu_item(self, item):
        # pylint: disable=attribute-defined-outside-init,no-member

        from ctypes import sizeof

        item_info = self.MENUITEMINFOW(cb_size=sizeof(self.MENUITEMINFOW))
        w_id = item["id"]
        text = item.get("text")
        is_checked = item.get("toggled")

        item_info.f_mask |= self.MIIM_ID
        item_info.w_id = w_id

        if text is not None:
            item_info.f_mask |= self.MIIM_STRING
            item_info.dw_type_data = text.replace("_", "&")  # Mnemonics use &
        else:
            item_info.f_type |= self.MFT_SEPARATOR

        if is_checked is not None:
            item_info.f_mask |= self.MIIM_STATE
            item_info.f_state |= self.MFS_CHECKED if is_checked else self.MFS_UNCHECKED

        return item_info

    def _show_menu(self):

        from ctypes import byref, windll, wintypes

        self._populate_menu()

        pos = wintypes.POINT()
        windll.user32.GetCursorPos(byref(pos))

        # PRB: Menus for Notification Icons Do Not Work Correctly
        # https://web.archive.org/web/20121015064650/http://support.microsoft.com/kb/135788

        windll.user32.SetForegroundWindow(self._h_wnd)
        windll.user32.TrackPopupMenu(
            self._menu,
            0,
            pos.x,
            pos.y,
            0,
            self._h_wnd,
            None
        )
        windll.user32.PostMessageW(self._h_wnd, self.WM_NULL, 0, 0)

    def _populate_menu(self):

        from ctypes import byref, windll

        self._destroy_menu()

        if self._menu is None:
            self._menu = windll.user32.CreatePopupMenu()

        for item in self.menu_items.values():
            if not item["visible"]:
                continue

            item_id = item["id"]
            item_info = self._serialize_menu_item(item)

            windll.user32.InsertMenuItemW(self._menu, item_id, False, byref(item_info))

    def _destroy_menu(self):

        if self._menu is None:
            return

        from ctypes import windll

        windll.user32.DestroyMenu(self._menu)
        self._menu = None

    def _set_icon_name(self, icon_name):
        self._update_notify_icon(icon_name=icon_name)

    def on_process_window_message(self, h_wnd, msg, w_param, l_param):

        from ctypes import windll, wintypes

        if msg == self.WM_TRAYICON:
            if l_param == self.WM_RBUTTONUP:
                # Icon pressed
                self._show_menu()

            elif l_param == self.WM_LBUTTONUP:
                # Icon pressed
                self.activate_callback()

            elif l_param in {self.NIN_BALLOONHIDE, self.NIN_BALLOONTIMEOUT, self.NIN_BALLOONUSERCLICK}:
                if l_param == self.NIN_BALLOONUSERCLICK and self._click_action is not None:
                    # Notification was clicked, perform action
                    self.application.lookup_action(self._click_action).activate(self._click_action_target)
                    self._click_action = self._click_action_target = None

                if not config.sections["ui"]["trayicon"]:
                    # Notification dismissed, but user has disabled tray icon
                    self._remove_notify_icon()

        elif msg == self.WM_COMMAND:
            # Menu item pressed
            menu_item_id = w_param & 0xFFFF
            menu_item_callback = self.menu_items[menu_item_id]["callback"]
            menu_item_callback()

        elif msg == self._wm_taskbarcreated:
            # Taskbar process restarted, create new icon
            self._remove_notify_icon()
            self._update_notify_icon()

        return windll.user32.DefWindowProcW(
            wintypes.HWND(h_wnd),
            msg,
            wintypes.WPARAM(w_param),
            wintypes.LPARAM(l_param)
        )

    def show_notification(self, title, message, action=None, action_target=None, high_priority=False):

        self._update_notify_icon(
            title=title, message=message, click_action=action, click_action_target=action_target,
            high_priority=high_priority)

    def unload(self, is_shutdown=True):

        self._remove_notify_icon()

        if not is_shutdown:
            # Keep notification support as long as we're running
            return

        self._destroy_h_icon()
        self._destroy_window()
        self._unregister_class()


class StatusIconImplementation(BaseImplementation):

    def __init__(self, application):

        super().__init__(application)

        if not hasattr(Gtk, "StatusIcon") or os.environ.get("WAYLAND_DISPLAY"):
            # GtkStatusIcon does not work on Wayland
            raise ImplementationUnavailable("StatusIcon implementation not available")

        self.tray_icon = Gtk.StatusIcon(tooltip_text=pynicotine.__application_name__)
        self.tray_icon.connect("activate", self.activate_callback)
        self.tray_icon.connect("popup-menu", self.on_status_icon_popup)

        self.gtk_menu = self._build_gtk_menu()
        GLib.idle_add(self._update_icon, priority=GLib.PRIORITY_HIGH_IDLE)

    def on_status_icon_popup(self, _status_icon, button, _activate_time):

        if button == 3:
            time = Gtk.get_current_event_time()
            self.gtk_menu.popup(None, None, None, None, button, time)

    def _set_item_text(self, item, text):
        super()._set_item_text(item, text)
        item["gtk_menu_item"].set_label(text)

    def _set_item_visible(self, item, visible):
        super()._set_item_visible(item, visible)
        item["gtk_menu_item"].set_visible(visible)

    def _set_item_toggled(self, item, toggled):

        super()._set_item_toggled(item, toggled)
        gtk_menu_item = item["gtk_menu_item"]

        with gtk_menu_item.handler_block(item["gtk_handler"]):
            gtk_menu_item.set_active(toggled)

    def _build_gtk_menu(self):

        gtk_menu = Gtk.Menu()

        for item in self.menu_items.values():
            text = item.get("text")

            if text is None:
                item["gtk_menu_item"] = gtk_menu_item = Gtk.SeparatorMenuItem(visible=True)
            else:
                gtk_menu_item_class = Gtk.CheckMenuItem if "toggled" in item else Gtk.MenuItem
                item["gtk_menu_item"] = gtk_menu_item = gtk_menu_item_class(
                    label=text, use_underline=True, visible=True
                )
                item["gtk_handler"] = gtk_menu_item.connect("activate", item["callback"])

            gtk_menu.append(gtk_menu_item)

        return gtk_menu

    def _set_icon_name(self, icon_name):

        if not self.tray_icon.get_visible():
            self.tray_icon.set_visible(True)

        if self.tray_icon.is_embedded() and self.tray_icon.get_icon_name() != icon_name:
            self.tray_icon.set_from_icon_name(icon_name)

    def unload(self, is_shutdown=True):
        if self.tray_icon.get_visible():
            self.tray_icon.set_visible(False)


class TrayIcon:

    def __init__(self, application):

        self.application = application
        self.available = True
        self.implementation = None
        self.watch_id = None

        self._watch_availability()
        self.load()

    def _watch_availability(self):

        if sys.platform in {"win32", "darwin"}:
            return

        if self.watch_id is not None:
            return

        self.watch_id = Gio.bus_watch_name(
            bus_type=Gio.BusType.SESSION,
            name="org.kde.StatusNotifierWatcher",
            flags=Gio.BusNameWatcherFlags.NONE,
            name_appeared_closure=self.load,
            name_vanished_closure=self.unload
        )

    def _unwatch_availability(self):

        if self.watch_id is not None:
            Gio.bus_unwatch_name(self.watch_id)
            self.watch_id = None

    def load(self, *_args):

        self.available = True

        if sys.platform == "win32":
            # Always keep tray icon loaded for Windows notification support
            pass

        elif not config.sections["ui"]["trayicon"]:
            # No need to have tray icon loaded now (unless this is Windows)
            return

        if self.implementation is None:
            if sys.platform == "win32":
                self.implementation = Win32Implementation(self.application)

            elif sys.platform == "darwin":
                self.available = False

            else:
                try:
                    self.implementation = StatusNotifierImplementation(self.application)

                except ImplementationUnavailable:
                    try:
                        self.implementation = StatusIconImplementation(self.application)
                        self._unwatch_availability()

                    except ImplementationUnavailable:
                        self.available = False
                        return

        self.update()

    def set_download_status(self, status):
        if self.implementation:
            self.implementation.set_download_status(status)

    def set_upload_status(self, status):
        if self.implementation:
            self.implementation.set_upload_status(status)

    def show_notification(self, title, message, action=None, action_target=None, high_priority=False):

        if self.implementation:
            self.implementation.show_notification(
                title=title, message=message, action=action, action_target=action_target,
                high_priority=high_priority)

    def update(self):
        if self.implementation:
            self.implementation.update()

    def unload(self, *_args, is_shutdown=True):

        if self.implementation:
            self.implementation.unload(is_shutdown=is_shutdown)
            self.implementation.is_visible = False

        if is_shutdown:
            self.implementation = None
            self.available = False

    def destroy(self):
        self.unload()
        self.__dict__.clear()



# COPYRIGHT (C) 2022-2024 Nicotine+ Contributors
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import time

from collections import deque

from gi.repository import GObject

from pynicotine.config import config
from pynicotine.core import core
from pynicotine.events import events
from pynicotine.gtkgui.application import GTK_API_VERSION
from pynicotine.gtkgui.widgets import ui
from pynicotine.gtkgui.widgets.accelerator import Accelerator
from pynicotine.gtkgui.widgets.popover import Popover
from pynicotine.gtkgui.widgets.textentry import CompletionEntry
from pynicotine.gtkgui.widgets.theme import USER_STATUS_ICON_NAMES
from pynicotine.gtkgui.widgets.theme import add_css_class
from pynicotine.gtkgui.widgets.treeview import TreeView
from pynicotine.logfacility import log
from pynicotine.slskmessages import UserStatus
from pynicotine.utils import encode_path


class ChatHistory(Popover):

    def __init__(self, window):

        (
            self.container,
            self.list_container,
            self.search_entry
        ) = ui.load(scope=self, path="popovers/chathistory.ui")

        super().__init__(
            window=window,
            content_box=self.container,
            width=1000,
            height=700
        )

        self.list_view = TreeView(
            window, parent=self.list_container, activate_row_callback=self.on_show_user,
            search_entry=self.search_entry,
            columns={
                "status": {
                    "column_type": "icon",
                    "title": _("Status"),
                    "width": 25,
                    "hide_header": True
                },
                "user": {
                    "column_type": "text",
                    "title": _("User"),
                    "width": 175,
                    "iterator_key": True
                },
                "latest_message": {
                    "column_type": "text",
                    "title": _("Latest Message")
                },

                # Hidden data columns
                "timestamp_data": {
                    "data_type": GObject.TYPE_UINT64,
                    "default_sort_type": "descending"
                }
            }
        )

        Accelerator("<Primary>f", self.widget, self.on_search_accelerator)
        self.completion_entry = CompletionEntry(window.private_entry, self.list_view.model, column=1)

        if GTK_API_VERSION >= 4:
            inner_button = next(iter(window.private_history_button))
            add_css_class(widget=inner_button, css_class="arrow-button")

        self.set_menu_button(window.private_history_button)
        self.load_users()

        for event_name, callback in (
            ("server-login", self.server_login),
            ("server-disconnect", self.server_disconnect),
            ("user-status", self.user_status)
        ):
            events.connect(event_name, callback)

    def destroy(self):

        self.list_view.destroy()
        self.completion_entry.destroy()

        super().destroy()

    def server_login(self, msg):

        if not msg.success:
            return

        for iterator in self.list_view.iterators.values():
            username = self.list_view.get_row_value(iterator, "user")
            core.users.watch_user(username, context="chathistory")

    def server_disconnect(self, *_args):
        for iterator in self.list_view.iterators.values():
            self.list_view.set_row_value(iterator, "status", USER_STATUS_ICON_NAMES[UserStatus.OFFLINE])

    @staticmethod
    def load_user(file_path):
        """Reads the username and latest message from a given log file path.

        Usernames are first extracted from the file name. In case the
        extracted username contains underscores, attempt to fetch the
        original username from logged messages, since illegal filename
        characters are substituted with underscores.
        """

        username = os.path.basename(file_path[:-4]).decode("utf-8", "replace")
        is_safe_username = ("_" not in username)
        login_username = config.sections["server"]["login"]
        timestamp = os.path.getmtime(file_path)

        read_num_lines = 1 if is_safe_username else 25
        latest_message = None

        with open(file_path, "rb") as lines:
            lines = deque(lines, read_num_lines)

            for line in lines:
                try:
                    line = line.decode("utf-8")

                except UnicodeDecodeError:
                    line = line.decode("latin-1")

                if latest_message is None:
                    latest_message = line

                    if is_safe_username:
                        break

                    username_chars = set(username.replace("_", ""))

                if login_username in line:
                    continue

                if " [" not in line or "] " not in line:
                    continue

                start = line.find(" [") + 2
                end = line.find("] ", start)
                line_username_len = (end - start)

                if len(username) != line_username_len:
                    continue

                line_username = line[start:end]

                if username == line_username:
                    # Nothing to do, username is already correct
                    break

                if username_chars.issubset(line_username):
                    username = line_username
                    break

        return username, latest_message, timestamp

    def load_users(self):

        self.list_view.freeze()

        try:
            with os.scandir(encode_path(log.private_chat_folder_path)) as entries:
                for entry in entries:
                    if not entry.is_file() or not entry.name.endswith(b".log"):
                        continue

                    try:
                        username, latest_message, timestamp = self.load_user(entry.path)

                    except OSError:
                        continue

                    if latest_message is not None:
                        self.update_user(username, latest_message.strip(), timestamp)

        except OSError:
            pass

        self.list_view.unfreeze()

    def remove_user(self, username):

        iterator = self.list_view.iterators.get(username)

        if iterator is not None:
            self.list_view.remove_row(iterator)

    def update_user(self, username, message, timestamp=None):

        self.remove_user(username)
        core.users.watch_user(username, context="chathistory")

        if not timestamp:
            timestamp_format = config.sections["logging"]["log_timestamp"]
            timestamp = time.time()
            h_timestamp = time.strftime(timestamp_format)
            message = f"{h_timestamp} {message}"

        status = core.users.statuses.get(username, UserStatus.OFFLINE)

        self.list_view.add_row([
            USER_STATUS_ICON_NAMES[status],
            username,
            message,
            int(timestamp)
        ], select_row=False)

    def user_status(self, msg):

        iterator = self.list_view.iterators.get(msg.user)

        if iterator is None:
            return

        status_icon_name = USER_STATUS_ICON_NAMES.get(msg.status)

        if status_icon_name and status_icon_name != self.list_view.get_row_value(iterator, "status"):
            self.list_view.set_row_value(iterator, "status", status_icon_name)

    def on_show_user(self, *_args):

        for iterator in self.list_view.get_selected_rows():
            username = self.list_view.get_row_value(iterator, "user")

            core.privatechat.show_user(username)
            self.close(use_transition=False)
            return

    def on_search_accelerator(self, *_args):
        """Ctrl+F - Search users."""

        self.search_entry.grab_focus()
        return True



# COPYRIGHT (C) 2022-2024 Nicotine+ Contributors
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pynicotine.pluginsystem import BasePlugin
from pynicotine.shares import PermissionLevel
from pynicotine.slskmessages import UserStatus


class _CommandGroup:
    CHAT = _("Chat")
    CHAT_ROOMS = _("Chat Rooms")
    PRIVATE_CHAT = _("Private Chat")
    NETWORK_FILTERS = _("Network Filters")
    SEARCH_FILES = _("Search Files")
    SHARES = _("Shares")
    USERS = _("Users")


class Plugin(BasePlugin):

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.commands = {
            "help": {
                "aliases": ["?"],
                "callback": self.help_command,
                "description": _("List available commands"),
                "parameters": ["[query]"]
            },
            "connect": {
                "callback": self.connect_command,
                "description": _("Connect to the server"),
            },
            "disconnect": {
                "callback": self.disconnect_command,
                "description": _("Disconnect from the server"),
            },
            "away": {
                "aliases": ["a"],
                "callback": self.away_command,
                "description": _("Toggle away status"),
            },
            "plugin": {
                "callback": self.plugin_handler_command,
                "description": _("Manage plugins"),
                "parameters": ["<toggle|reload|info>", "<plugin name>"]
            },
            "quit": {
                "aliases": ["q", "exit"],
                "callback": self.quit_command,
                "description": _("Quit Nicotine+"),
                "parameters": ["[force]"]
            },
            "clear": {
                "aliases": ["cl"],
                "callback": self.clear_command,
                "description": _("Clear chat window"),
                "disable": ["cli"],
                "group": _CommandGroup.CHAT,
            },
            "me": {
                "callback": self.me_command,
                "description": _("Say something in the third-person"),
                "disable": ["cli"],
                "group": _CommandGroup.CHAT,
                "parameters": ["<something..>"]
            },
            "now": {
                "callback": self.now_command,
                "description": _("Announce the song currently playing"),
                "disable": ["cli"],
                "group": _CommandGroup.CHAT
            },
            "join": {
                "aliases": ["j"],
                "callback": self.join_command,
                "description": _("Join chat room"),
                "disable": ["cli"],
                "group": _CommandGroup.CHAT_ROOMS,
                "parameters": ["<room>"]
            },
            "leave": {
                "aliases": ["l"],
                "callback": self.leave_command,
                "description": _("Leave chat room"),
                "disable": ["cli"],
                "group": _CommandGroup.CHAT_ROOMS,
                "parameters": ["<room>"],
                "parameters_chatroom": ["[room]"]
            },
            "say": {
                "callback": self.say_command,
                "description": _("Say message in specified chat room"),
                "disable": ["cli"],
                "group": _CommandGroup.CHAT_ROOMS,
                "parameters": ["<room>", "<message..>"]
            },
            "pm": {
                "callback": self.pm_command,
                "description": _("Open private chat"),
                "disable": ["cli"],
                "group": _CommandGroup.PRIVATE_CHAT,
                "parameters": ["<user>"]
            },
            "close": {
                "aliases": ["c"],
                "callback": self.close_command,
                "description": _("Close private chat"),
                "disable": ["cli"],
                "group": _CommandGroup.PRIVATE_CHAT,
                "parameters_chatroom": ["<user>"],
                "parameters_private_chat": ["[user]"]
            },
            "ctcpversion": {
                "callback": self.ctcpversion_command,
                "description": _("Request user's client version"),
                "disable": ["cli"],
                "group": _CommandGroup.PRIVATE_CHAT,
                "parameters_chatroom": ["<user>"],
                "parameters_private_chat": ["[user]"]
            },
            "msg": {
                "aliases": ["m"],
                "callback": self.msg_command,
                "description": _("Send private message to user"),
                "disable": ["cli"],
                "group": _CommandGroup.PRIVATE_CHAT,
                "parameters": ["<user>", "<message..>"]
            },
            "add": {
                "aliases": ["buddy"],
                "callback": self.add_buddy_command,
                "description": _("Add user to buddy list"),
                "group": _CommandGroup.USERS,
                "parameters": ["<user>"],
                "parameters_private_chat": ["[user]"]
            },
            "rem": {
                "aliases": ["unbuddy"],
                "callback": self.remove_buddy_command,
                "description": _("Remove buddy from buddy list"),
                "group": _CommandGroup.USERS,
                "parameters": ["<user>"],
                "parameters_private_chat": ["[user]"]
            },
            "browse": {
                "aliases": ["b"],
                "callback": self.browse_user_command,
                "description": _("Browse files of user"),
                "disable": ["cli"],
                "group": _CommandGroup.USERS,
                "parameters": ["<user>"],
                "parameters_private_chat": ["[user]"]
            },
            "whois": {
                "aliases": ["info", "w"],
                "callback": self.whois_command,
                "description": _("Show user profile information"),
                "disable": ["cli"],
                "group": _CommandGroup.USERS,
                "parameters": ["<user>"],
                "parameters_private_chat": ["[user]"]
            },
            "ip": {
                "callback": self.ip_address_command,
                "description": _("Show IP address or username"),
                "group": _CommandGroup.NETWORK_FILTERS,
                "parameters": ["<user or ip>"],
                "parameters_private_chat": ["[user or ip]"]
            },
            "ban": {
                "callback": self.ban_command,
                "description": _("Block connections from user or IP address"),
                "group": _CommandGroup.NETWORK_FILTERS,
                "parameters": ["<user or ip>"],
                "parameters_private_chat": ["[user or ip]"]
            },
            "unban": {
                "callback": self.unban_command,
                "description": _("Remove user or IP address from ban lists"),
                "group": _CommandGroup.NETWORK_FILTERS,
                "parameters": ["<user or ip>"],
                "parameters_private_chat": ["[user or ip]"]
            },
            "ignore": {
                "callback": self.ignore_command,
                "description": _("Silence messages from user or IP address"),
                "disable": ["cli"],
                "group": _CommandGroup.NETWORK_FILTERS,
                "parameters": ["<user or ip>"],
                "parameters_private_chat": ["[user or ip]"]
            },
            "unignore": {
                "callback": self.unignore_command,
                "description": _("Remove user or IP address from ignore lists"),
                "disable": ["cli"],
                "group": _CommandGroup.NETWORK_FILTERS,
                "parameters": ["<user or ip>"],
                "parameters_private_chat": ["[user or ip]"]
            },
            "share": {
                "callback": self.share_command,
                "description": _("Add share"),
                "group": _CommandGroup.SHARES,
                "parameters": ["<public|buddy|trusted>", "<folder path>"]
            },
            "unshare": {
                "callback": self.unshare_command,
                "description": _("Remove share"),
                "group": _CommandGroup.SHARES,
                "parameters": ["<virtual name or folder path>"]
            },
            "shares": {
                "aliases": ["ls"],
                "callback": self.list_shares_command,
                "description": _("List shares"),
                "group": _CommandGroup.SHARES,
                "parameters": ["[public|buddy|trusted]"]
            },
            "rescan": {
                "callback": self.rescan_command,
                "description": _("Rescan shares"),
                "group": _CommandGroup.SHARES,
                "parameters": ["[force|rebuild]"]
            },
            "search": {
                "aliases": ["s"],
                "callback": self.search_command,
                "description": _("Start global file search"),
                "disable": ["cli"],
                "group": _CommandGroup.SEARCH_FILES,
                "parameters": ["<query>"]
            },
            "rsearch": {
                "aliases": ["rs"],
                "callback": self.search_rooms_command,
                "description": _("Search files in joined rooms"),
                "disable": ["cli"],
                "group": _CommandGroup.SEARCH_FILES,
                "parameters": ["<query>"]
            },
            "bsearch": {
                "aliases": ["bs"],
                "callback": self.search_buddies_command,
                "description": _("Search files of all buddies"),
                "disable": ["cli"],
                "group": _CommandGroup.SEARCH_FILES,
                "parameters": ["<query>"]
            },
            "usearch": {
                "aliases": ["us"],
                "callback": self.search_user_command,
                "description": _("Search a user's shared files"),
                "disable": ["cli"],
                "group": _CommandGroup.SEARCH_FILES,
                "parameters": ["<user>", "<query>"]
            }
        }

    # Application Commands #

    def help_command(self, args, user=None, room=None):

        if user is not None:
            command_interface = "private_chat"

        elif room is not None:
            command_interface = "chatroom"

        else:
            command_interface = "cli"

        search_query = " ".join(args.lower().split(" ", maxsplit=1))
        command_groups = self.parent.get_command_groups_data(command_interface, search_query=search_query)
        num_commands = sum(len(command_groups[x]) for x in command_groups)
        output_text = ""

        if not search_query:
            output_text += _("Listing %(num)i available commands:") % {"num": num_commands}
        else:
            output_text += _('Listing %(num)i available commands matching "%(query)s":') % {
                "num": num_commands,
                "query": search_query
            }

        for group_name, command_data in command_groups.items():
            output_text += f"\n\n{group_name}:"

            for command, aliases, parameters, description in command_data:
                command_message = f"/{', /'.join([command] + aliases)} {' '.join(parameters)}".strip()
                output_text += f"\n\t{command_message}  -  {description}"

        if not search_query:
            output_text += "\n\n" + _("Type %(command)s to list similar commands") % {"command": "/help [query]"}

        elif not num_commands:
            output_text += "\n" + _("Type %(command)s to list available commands") % {"command": "/help"}

        self.output(output_text)

    def connect_command(self, _args, **_unused):
        if self.core.users.login_status == UserStatus.OFFLINE:
            self.core.connect()

    def disconnect_command(self, _args, **_unused):
        if self.core.users.login_status != UserStatus.OFFLINE:
            self.core.disconnect()

    def away_command(self, _args, **_unused):

        if self.core.users.login_status == UserStatus.OFFLINE:
            self.output(_("%(user)s is offline") % {"user": self.config.sections["server"]["login"]})
            return

        self.core.users.set_away_mode(self.core.users.login_status != UserStatus.AWAY, save_state=True)

        if self.core.users.login_status == UserStatus.ONLINE:
            self.output(_("%(user)s is online") % {"user": self.core.users.login_username})
        else:
            self.output(_("%(user)s is away") % {"user": self.core.users.login_username})

    def quit_command(self, args, **_unused):

        force = (args.lstrip("-") in {"force", "f"})

        if force:
            self.core.quit()
        else:
            self.core.confirm_quit()

    # Chat #

    def clear_command(self, _args, user=None, room=None):

        if room is not None:
            self.core.chatrooms.clear_room_messages(room)

        elif user is not None:
            self.core.privatechat.clear_private_messages(user)

    def me_command(self, args, **_unused):
        self.send_message("/me " + args)  # /me is sent as plain text

    def now_command(self, _args, **_unused):
        self.core.now_playing.display_now_playing(callback=self.send_message)

    # Chat Rooms #

    def join_command(self, args, **_unused):
        room = self.core.chatrooms.sanitize_room_name(args)
        self.core.chatrooms.show_room(room)

    def leave_command(self, args, room=None, **_unused):

        if args:
            room = args

        if room not in self.core.chatrooms.joined_rooms:
            self.output(_("Not joined in room %s") % room)
            return False

        self.core.chatrooms.remove_room(room)
        return True

    def say_command(self, args, **_unused):

        room, text = args.split(maxsplit=1)

        if room not in self.core.chatrooms.joined_rooms:
            self.output(_("Not joined in room %s") % room)
            return False

        self.send_public(room, text)
        return True

    # Private Chat #

    def pm_command(self, args, **_unused):
        self.core.privatechat.show_user(args)

    def close_command(self, args, user=None, **_unused):

        if args:
            user = args

        if user not in self.core.privatechat.users:
            self.output(_("Not messaging with user %s") % user)
            return False

        self.core.privatechat.remove_user(user)
        self.output(_("Closed private chat of user %s") % user)
        return True

    def ctcpversion_command(self, args, user=None, **_unused):

        if args:
            user = args

        self.send_private(user, self.core.privatechat.CTCP_VERSION, show_ui=True)

    def msg_command(self, args, **_unused):
        user, text = args.split(maxsplit=1)
        self.send_private(user, text, show_ui=True, switch_page=False)

    # Users #

    def add_buddy_command(self, args, user=None, **_unused):

        if args:
            user = args

        self.core.buddies.add_buddy(user)

    def remove_buddy_command(self, args, user=None, **_unused):

        if args:
            user = args

        self.core.buddies.remove_buddy(user)

    def browse_user_command(self, args, user=None, **_unused):

        if args:
            user = args

        self.core.userbrowse.browse_user(user)

    def whois_command(self, args, user=None, **_unused):

        if args:
            user = args

        self.core.userinfo.show_user(user)

    # Network Filters #

    def ip_address_command(self, args, user=None, **_unused):

        if self.core.network_filter.is_ip_address(args):
            self.output(self.core.network_filter.get_online_username(args))
            return

        if args:
            user = args

        self.core.users.request_ip_address(user, notify=True)

    def ban_command(self, args, user=None, **_unused):

        if self.core.network_filter.is_ip_address(args):
            banned_ip_address = self.core.network_filter.ban_user_ip(ip_address=args)
        else:
            if args:
                user = args

            banned_ip_address = None
            self.core.network_filter.ban_user(user)

        self.output(_("Banned %s") % (banned_ip_address or user))

    def unban_command(self, args, user=None, **_unused):

        if self.core.network_filter.is_ip_address(args):
            unbanned_ip_addresses = self.core.network_filter.unban_user_ip(ip_address=args)
            self.core.network_filter.unban_user(self.core.network_filter.get_online_username(args))
        else:
            if args:
                user = args

            unbanned_ip_addresses = self.core.network_filter.unban_user_ip(user)
            self.core.network_filter.unban_user(user)

        self.output(_("Unbanned %s") % (" & ".join(unbanned_ip_addresses) or user))

    def ignore_command(self, args, user=None, **_unused):

        if self.core.network_filter.is_ip_address(args):
            ignored_ip_address = self.core.network_filter.ignore_user_ip(ip_address=args)
        else:
            if args:
                user = args

            ignored_ip_address = None
            self.core.network_filter.ignore_user(user)

        self.output(_("Ignored %s") % (ignored_ip_address or user))

    def unignore_command(self, args, user=None, **_unused):

        if self.core.network_filter.is_ip_address(args):
            unignored_ip_addresses = self.core.network_filter.unignore_user_ip(ip_address=args)
            self.core.network_filter.unignore_user(self.core.network_filter.get_online_username(args))
        else:
            if args:
                user = args

            unignored_ip_addresses = self.core.network_filter.unignore_user_ip(user)
            self.core.network_filter.unignore_user(user)

        self.output(_("Unignored %s") % (" & ".join(unignored_ip_addresses) or user))

    # Configure Shares #

    def rescan_command(self, args, **_unused):

        rebuild = ("rebuild" in args)
        force = ("force" in args) or rebuild

        self.core.shares.rescan_shares(rebuild=rebuild, force=force)

    def list_shares_command(self, args, **_unused):

        permission_levels = {
            0: PermissionLevel.PUBLIC,
            1: PermissionLevel.BUDDY,
            2: PermissionLevel.TRUSTED
        }
        share_groups = self.core.shares.get_shared_folders()
        num_total = num_listed = 0

        for group_index, share_group in enumerate(share_groups):
            permission_level = permission_levels.get(group_index)
            num_shares = len(share_group)
            num_total += num_shares

            if not num_shares or args and permission_level not in args.lower():
                continue

            self.output("\n" + f"{num_shares} {permission_level} shares:")

            for virtual_name, folder_path, *_ignored in share_group:
                self.output(f'• "{virtual_name}" {folder_path}')

            num_listed += num_shares

        self.output("\n" + _("%(num_listed)s shares listed (%(num_total)s configured)") % {
            "num_listed": num_listed,
            "num_total": num_total
        })

    def share_command(self, args, **_unused):

        permission_level, folder_path = args.split(maxsplit=1)
        folder_path = folder_path.strip(' "')
        virtual_name = self.core.shares.add_share(folder_path, permission_level=permission_level)

        if not virtual_name:
            self.output(_("Cannot share inaccessible folder \"%s\"") % folder_path)
            return False

        self.output(_("Added %(group_name)s share \"%(virtual_name)s\" (rescan required)") % {
            "group_name": permission_level,
            "virtual_name": virtual_name
        })
        return True

    def unshare_command(self, args, **_unused):

        virtual_name_or_folder_path = args.strip(' "')

        if not self.core.shares.remove_share(virtual_name_or_folder_path):
            self.output(_("No share with name \"%s\"") % virtual_name_or_folder_path)
            return False

        self.output(_("Removed share \"%s\" (rescan required)") % virtual_name_or_folder_path)
        return True

    # Search Files #

    def search_command(self, args, **_unused):
        self.core.search.do_search(args, "global")

    def search_rooms_command(self, args, **_unused):
        self.core.search.do_search(args, "rooms")

    def search_buddies_command(self, args, **_unused):
        self.core.search.do_search(args, "buddies")

    def search_user_command(self, args, **_unused):
        user, query = args.split(maxsplit=1)
        self.core.search.do_search(query, "user", users=[user])

    # Plugin Commands #

    def plugin_handler_command(self, args, **_unused):

        action, plugin_name = args.split(maxsplit=1)

        if action == "toggle":
            self.parent.toggle_plugin(plugin_name)

        elif action == "reload":
            self.parent.reload_plugin(plugin_name)

        elif action == "info":
            plugin_info = self.parent.get_plugin_info(plugin_name)

            for key, value in plugin_info.items():
                self.output(f"• {key}: {value}")



# COPYRIGHT (C) 2022-2025 Nicotine+ Contributors
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import time

from collections import deque
from getpass import getpass
from threading import Thread

from pynicotine.events import events
from pynicotine.logfacility import log


class CLIInputProcessor(Thread):
    __slots__ = ("has_custom_prompt", "prompt_message", "prompt_callback", "prompt_silent")

    def __init__(self):

        super().__init__(name="CLIInputProcessor", daemon=True)

        try:
            # Enable line editing and history
            import readline  # noqa: F401  # pylint:disable=unused-import

        except ImportError:
            # Readline is not available on this OS
            pass

        self.has_custom_prompt = False
        self.prompt_message = ""
        self.prompt_callback = None
        self.prompt_silent = False

    def run(self):

        while True:
            # Small time window to set custom prompt
            time.sleep(0.25)

            try:
                self._handle_prompt()

            except Exception as error:
                log.add_debug("CLI input prompt is no longer available: %s", error)
                return

    def _handle_prompt_callback(self, user_input, callback):

        if not callback:
            return False

        events.invoke_main_thread(callback, user_input)
        return True

    def _handle_prompt_command(self, user_input):

        if not user_input:
            return False

        command, _separator, args = user_input.strip().partition(" ")
        args = args.strip()

        if command.startswith("/"):
            command = command[1:]

        events.emit_main_thread("cli-command", command, args)
        return True

    def _handle_prompt(self):

        callback = self.prompt_callback
        input_func = getpass if self.prompt_silent else input
        self.has_custom_prompt = (callback is not None)

        user_input = input_func(self.prompt_message)

        self.has_custom_prompt = False
        self.prompt_message = ""
        self.prompt_callback = None
        self.prompt_silent = False

        events.emit("cli-prompt-finished")

        # Check if custom prompt is active
        if self._handle_prompt_callback(user_input, callback):
            return

        # No custom prompt, treat input as command
        self._handle_prompt_command(user_input)


class CLI:
    __slots__ = ("_input_processor", "_log_message_queue", "_tty_attributes")

    def __init__(self):

        self._input_processor = CLIInputProcessor()
        self._log_message_queue = deque(maxlen=1000)
        self._tty_attributes = None

        events.connect("quit", self._quit)

    def enable_prompt(self):

        try:
            import termios  # pylint: disable=import-error
            self._tty_attributes = termios.tcgetattr(sys.stdin)

        except Exception:
            # Not a terminal, or using Windows
            pass

        self._input_processor.start()

    def enable_logging(self):

        for event_name, callback in (
            ("cli-prompt-finished", self._cli_prompt_finished),
            ("log-message", self._log_message)
        ):
            events.connect(event_name, callback)

    def prompt(self, message, callback, is_silent=False):

        self._input_processor.prompt_message = message
        self._input_processor.prompt_callback = callback
        self._input_processor.prompt_silent = is_silent

    def _print_log_message(self, log_message):

        try:
            print(log_message, flush=True)

        except OSError:
            # stdout is gone, prevent future errors
            sys.stdout = open(os.devnull, "w", encoding="utf-8")  # pylint: disable=consider-using-with

    def _cli_prompt_finished(self):
        while self._log_message_queue:
            self._print_log_message(self._log_message_queue.popleft())

    def _log_message(self, timestamp_format, msg, _title, _level):

        if timestamp_format:
            timestamp = time.strftime(timestamp_format)
            log_message = f"[{timestamp}] {msg}"
        else:
            log_message = msg

        if self._input_processor.has_custom_prompt:
            # Don't print log messages while custom prompt is active
            self._log_message_queue.append(log_message)
            return

        self._print_log_message(log_message)

    def _quit(self):
        """Restores TTY attributes and re-enables echo on quit."""

        if self._tty_attributes is None:
            return

        import termios  # pylint: disable=import-error

        try:
            termios.tcsetattr(sys.stdin, termios.TCSANOW, self._tty_attributes)

        except termios.error:
            # stdin is gone
            pass

        self._tty_attributes = None


cli = CLI()



# COPYRIGHT (C) 2021-2024 Nicotine+ Contributors
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

from gi.repository import Gdk
from gi.repository import Gtk

from pynicotine.gtkgui.application import GTK_API_VERSION


class Accelerator:

    if GTK_API_VERSION >= 4:
        shortcut_triggers = {}
    else:
        KEYMAP = Gdk.Keymap.get_for_display(Gdk.Display.get_default())
        keycodes_mods = {}

    def __init__(self, accelerator, widget, callback, user_data=None):

        if GTK_API_VERSION >= 4:
            if sys.platform == "darwin":
                # Use Command key instead of Ctrl in accelerators on macOS
                accelerator = accelerator.replace("<Primary>", "<Meta>")

            shortcut_trigger = self.shortcut_triggers.get(accelerator)

            if not hasattr(widget, "shortcut_controller"):
                widget.shortcut_controller = Gtk.ShortcutController(
                    propagation_phase=Gtk.PropagationPhase.CAPTURE
                )
                widget.add_controller(widget.shortcut_controller)

            if not shortcut_trigger:
                self.shortcut_triggers[accelerator] = shortcut_trigger = Gtk.ShortcutTrigger.parse_string(accelerator)

            widget.shortcut_controller.add_shortcut(
                Gtk.Shortcut(
                    trigger=shortcut_trigger,
                    action=Gtk.CallbackAction.new(callback, user_data)
                )
            )
            return

        # GTK 3 replacement for Gtk.ShortcutController
        self.keycodes, self.required_mods = self.parse_accelerator(accelerator)
        self.callback = callback
        self.user_data = user_data

        widget.connect("key-press-event", self._activate_accelerator)

    @classmethod
    def parse_accelerator(cls, accelerator):

        keycodes_mods_accel = cls.keycodes_mods.get(accelerator)

        if not keycodes_mods_accel:
            *_args, key, mods = Gtk.accelerator_parse(accelerator)

            if key:
                _valid, keys = cls.KEYMAP.get_entries_for_keyval(key)
                keycodes = {key.keycode for key in keys}
            else:
                keycodes = []

            cls.keycodes_mods[accelerator] = keycodes_mods_accel = (keycodes, mods)

        return keycodes_mods_accel

    def _activate_accelerator(self, widget, event):

        activated_mods = event.state
        required_mods = self.required_mods
        excluded_mods = ALL_MODIFIERS & ~required_mods

        if required_mods & ~activated_mods:
            # Missing required modifiers
            return False

        if activated_mods & excluded_mods:
            # Too many/irrelevant modifiers
            return False

        if event.hardware_keycode not in self.keycodes:
            # Invalid key
            return False

        return self.callback(widget, None, self.user_data)


if GTK_API_VERSION == 3:
    ALL_MODIFIERS = (Accelerator.parse_accelerator("<Primary>")[1]
                     | Accelerator.parse_accelerator("<Shift>")[1]
                     | Accelerator.parse_accelerator("<Alt>")[1])
else:
    ALL_MODIFIERS = []



# COPYRIGHT (C) 2020-2024 Nicotine+ Contributors
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time

from pynicotine.core import core
from pynicotine.events import events
from pynicotine.gtkgui.widgets import ui
from pynicotine.gtkgui.widgets.dialogs import Dialog
from pynicotine.gtkgui.widgets.dialogs import OptionDialog
from pynicotine.utils import human_size
from pynicotine.utils import humanize


class Statistics(Dialog):

    def __init__(self, application):

        (
            self.completed_downloads_session_label,
            self.completed_downloads_total_label,
            self.completed_uploads_session_label,
            self.completed_uploads_total_label,
            self.container,
            self.current_session_label,
            self.downloaded_size_session_label,
            self.downloaded_size_total_label,
            self.reset_button,
            self.since_timestamp_total_label,
            self.uploaded_size_session_label,
            self.uploaded_size_total_label
        ) = ui.load(scope=self, path="dialogs/statistics.ui")

        self.stat_id_labels = {
            "completed_downloads": {
                "session": self.completed_downloads_session_label,
                "total": self.completed_downloads_total_label
            },
            "completed_uploads": {
                "session": self.completed_uploads_session_label,
                "total": self.completed_uploads_total_label
            },
            "downloaded_size": {
                "session": self.downloaded_size_session_label,
                "total": self.downloaded_size_total_label
            },
            "uploaded_size": {
                "session": self.uploaded_size_session_label,
                "total": self.uploaded_size_total_label
            },
            "since_timestamp": {
                "total": self.since_timestamp_total_label
            }
        }

        super().__init__(
            parent=application.window,
            content_box=self.container,
            show_callback=self.on_show,
            title=_("Transfer Statistics"),
            width=425
        )

        events.connect("update-stat", self.update_stat)

    def update_stat(self, stat_id, session_value, total_value):

        current_stat_id_labels = self.stat_id_labels.get(stat_id)

        if not current_stat_id_labels:
            return

        if not self.widget.get_visible():
            return

        if stat_id in {"downloaded_size", "uploaded_size"}:
            session_value = human_size(session_value)
            total_value = human_size(total_value)

        elif stat_id == "since_timestamp":
            session_value = None
            total_value = (_("Total Since %(date)s") % {
                "date": time.strftime("%x", time.localtime(total_value))} if total_value > 0 else None)

        else:
            session_value = humanize(session_value)
            total_value = humanize(total_value)

        if session_value is not None:
            session_label = current_stat_id_labels["session"]

            if session_label.get_text() != session_value:
                session_label.set_text(session_value)

        if total_value is not None:
            total_label = current_stat_id_labels["total"]

            if total_label.get_text() != total_value:
                total_label.set_text(total_value)

    def on_reset_statistics_response(self, *_args):
        core.statistics.reset_stats()

    def on_reset_statistics(self, *_args):

        OptionDialog(
            parent=self,
            title=_("Reset Transfer Statistics?"),
            message=_("Do you really want to reset transfer statistics?"),
            destructive_response_id="ok",
            callback=self.on_reset_statistics_response
        ).present()

    def on_close(self, *_args):
        self.close()

    def on_show(self, *_args):
        core.statistics.update_stats()



# COPYRIGHT (C) 2020-2024 Nicotine+ Contributors
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pynicotine
from pynicotine.config import config
from pynicotine.core import core
from pynicotine.events import events
from pynicotine.logfacility import log
from pynicotine.slskmessages import MessageAcked
from pynicotine.slskmessages import MessageUser
from pynicotine.slskmessages import MessageUsers
from pynicotine.slskmessages import SayChatroom
from pynicotine.slskmessages import UserStatus
from pynicotine.utils import censor_text
from pynicotine.utils import find_whole_word


class PrivateChat:
    __slots__ = ("completions", "private_message_queue", "away_message_users", "users")

    CTCP_VERSION = "\x01VERSION\x01"

    def __init__(self):

        self.completions = set()
        self.private_message_queue = {}
        self.away_message_users = set()
        self.users = set()

        for event_name, callback in (
            ("message-user", self._message_user),
            ("peer-address", self._get_peer_address),
            ("quit", self._quit),
            ("server-login", self._server_login),
            ("server-disconnect", self._server_disconnect),
            ("start", self._start),
            ("user-status", self._user_status)
        ):
            events.connect(event_name, callback)

    def _start(self):

        if not config.sections["privatechat"]["store"]:
            # Clear list of previously open chats if we don't want to restore them
            config.sections["privatechat"]["users"].clear()
            return

        for username in config.sections["privatechat"]["users"]:
            if isinstance(username, str) and username not in self.users:
                self.show_user(username, switch_page=False, remembered=True)

        self.update_completions()

    def _quit(self):
        self.remove_all_users(is_permanent=False)
        self.completions.clear()

    def _server_login(self, msg):

        if not msg.success:
            return

        for username in self.users:
            core.users.watch_user(username, context="privatechat")  # Get notified of user status

    def _server_disconnect(self, _msg):

        self.private_message_queue.clear()
        self.away_message_users.clear()
        self.update_completions()

    def add_user(self, username):

        if username in self.users:
            return

        self.users.add(username)

        if username not in config.sections["privatechat"]["users"]:
            config.sections["privatechat"]["users"].insert(0, username)

    def remove_user(self, username, is_permanent=True):

        if is_permanent and username in config.sections["privatechat"]["users"]:
            config.sections["privatechat"]["users"].remove(username)

        self.users.remove(username)
        core.users.unwatch_user(username, context="privatechat")
        events.emit("private-chat-remove-user", username)

    def remove_all_users(self, is_permanent=True):
        for username in self.users.copy():
            self.remove_user(username, is_permanent)

    def show_user(self, username, switch_page=True, remembered=False):

        self.add_user(username)
        events.emit("private-chat-show-user", username, switch_page, remembered)
        core.users.watch_user(username, context="privatechat")

    def clear_private_messages(self, username):
        events.emit("clear-private-messages", username)

    def private_message_queue_add(self, msg):
        """Queue a private message until we've received a user's IP address."""

        username = msg.user

        if username not in self.private_message_queue:
            self.private_message_queue[username] = [msg]
        else:
            self.private_message_queue[username].append(msg)

    def send_automatic_message(self, username, message):
        self.send_message(username, f"[Automatic Message] {message}")

    def echo_message(self, username, message, message_type="local"):
        events.emit("echo-private-message", username, message, message_type)

    def send_message(self, username, message):

        user_text = core.pluginhandler.outgoing_private_chat_event(username, message)
        if user_text is None:
            return

        username, message = user_text

        if config.sections["words"]["replacewords"] and message != self.CTCP_VERSION:
            for word, replacement in config.sections["words"]["autoreplaced"].items():
                message = message.replace(str(word), str(replacement))

        core.send_message_to_server(MessageUser(username, message))
        core.pluginhandler.outgoing_private_chat_notification(username, message)

        events.emit("message-user", MessageUser(username, message))

    def send_message_users(self, target, message):

        if not message:
            return

        users = None

        if target == "buddies":
            users = set(core.buddies.users)

        elif target == "downloading":
            users = core.uploads.get_downloading_users()

        if users:
            core.send_message_to_server(MessageUsers(users, message))

    def _get_peer_address(self, msg):
        """Server code 3.

        Received a user's IP address, process any queued private
        messages and check if the IP is ignored
        """

        username = msg.user

        if username not in self.private_message_queue:
            return

        for queued_msg in self.private_message_queue[username][:]:
            self.private_message_queue[username].remove(queued_msg)
            queued_msg.user = username
            events.emit("message-user", queued_msg, queued_message=True)

    def _user_status(self, msg):
        """Server code 7."""

        if msg.user == core.users.login_username and msg.status != UserStatus.AWAY:
            # Reset list of users we've sent away messages to when the away session ends
            self.away_message_users.clear()

        if msg.status == UserStatus.OFFLINE:
            self.private_message_queue.pop(msg.user, None)

    def get_message_type(self, text, is_outgoing_message):

        if text.startswith("/me "):
            return "action"

        if is_outgoing_message:
            return "local"

        if core.users.login_username and find_whole_word(core.users.login_username.lower(), text.lower()) > -1:
            return "hilite"

        return "remote"

    def _message_user(self, msg, queued_message=False):
        """Server code 22."""

        is_outgoing_message = (msg.message_id is None)

        username = msg.user
        tag_username = (core.users.login_username if is_outgoing_message else username)
        message = msg.message
        timestamp = msg.timestamp if not msg.is_new_message else None

        if not is_outgoing_message:
            if not queued_message:
                log.add_chat(_("Private message from user '%(user)s': %(message)s"), {
                    "user": username,
                    "message": message
                })

                core.send_message_to_server(MessageAcked(msg.message_id))

            if username == "server":
                start_str = "The room you are trying to enter ("

                if message.startswith(start_str) and ") " in message:
                    # Redirect message to chat room tab if join wasn't successful
                    msg.user = None
                    room = message[len(start_str):message.rfind(") ")]
                    events.emit("say-chat-room", SayChatroom(room=room, message=message, user=username))
                    return
            else:
                # Check ignore status for all other users except "server"
                if core.network_filter.is_user_ignored(username):
                    msg.user = None
                    return

                user_address = core.users.addresses.get(username)

                if user_address is not None:
                    if core.network_filter.is_user_ip_ignored(username):
                        msg.user = None
                        return

                elif not queued_message:
                    # Ask for user's IP address and queue the private message until we receive the address
                    if username not in self.private_message_queue:
                        core.users.request_ip_address(username)

                    self.private_message_queue_add(msg)
                    msg.user = None
                    return

            user_text = core.pluginhandler.incoming_private_chat_event(username, message)
            if user_text is None:
                msg.user = None
                return

            self.show_user(username, switch_page=False)

            _username, msg.message = user_text
            message = msg.message

        msg.message_type = self.get_message_type(message, is_outgoing_message)
        is_action_message = (msg.message_type == "action")
        is_ctcp_version = (message == self.CTCP_VERSION)

        # SEND CLIENT VERSION to user if the following string is sent
        if is_ctcp_version:
            msg.message = message = "CTCP VERSION"

        if is_action_message:
            message = message.replace("/me ", "", 1)

        if not is_outgoing_message and config.sections["words"]["censorwords"]:
            message = censor_text(message, censored_patterns=config.sections["words"]["censored"])

        if is_action_message:
            msg.formatted_message = msg.message = f"* {tag_username} {message}"
        else:
            msg.formatted_message = f"[{tag_username}] {message}"

        if config.sections["logging"]["privatechat"] or username in config.sections["logging"]["private_chats"]:
            log.write_log_file(
                folder_path=log.private_chat_folder_path,
                basename=username, text=msg.formatted_message, timestamp=timestamp
            )

        if is_outgoing_message:
            return

        core.pluginhandler.incoming_private_chat_notification(username, msg.message)

        if is_ctcp_version and not config.sections["server"]["ctcpmsgs"]:
            self.send_message(username, f"{pynicotine.__application_name__} {pynicotine.__version__}")

        if not msg.is_new_message:
            # Message was sent while offline, don't auto-reply
            return

        autoreply = config.sections["server"]["autoreply"]

        if (autoreply and core.users.login_status == UserStatus.AWAY
                and username not in self.away_message_users):
            self.send_automatic_message(username, autoreply)
            self.away_message_users.add(username)

    def update_completions(self):

        self.completions.clear()
        self.completions.add(config.sections["server"]["login"])

        if config.sections["words"]["roomnames"]:
            self.completions.update(core.chatrooms.server_rooms)

        if config.sections["words"]["buddies"]:
            self.completions.update(core.buddies.users)

        if config.sections["words"]["commands"]:
            self.completions.update(core.pluginhandler.get_command_list("private_chat"))

        events.emit("private-chat-completions", self.completions.copy())



# COPYRIGHT (C) 2020-2024 Nicotine+ Contributors
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pynicotine.config import config
from pynicotine.core import core
from pynicotine.events import events
from pynicotine.logfacility import log
from pynicotine.slskmessages import JoinGlobalRoom
from pynicotine.slskmessages import JoinRoom
from pynicotine.slskmessages import LeaveGlobalRoom
from pynicotine.slskmessages import LeaveRoom
from pynicotine.slskmessages import PrivateRoomAddOperator
from pynicotine.slskmessages import PrivateRoomAddUser
from pynicotine.slskmessages import PrivateRoomCancelMembership
from pynicotine.slskmessages import PrivateRoomDisown
from pynicotine.slskmessages import PrivateRoomRemoveOperator
from pynicotine.slskmessages import PrivateRoomRemoveUser
from pynicotine.slskmessages import PrivateRoomToggle
from pynicotine.slskmessages import RoomList
from pynicotine.slskmessages import RoomTickerSet
from pynicotine.slskmessages import SayChatroom
from pynicotine.utils import censor_text
from pynicotine.utils import find_whole_word


class JoinedRoom:
    __slots__ = ("name", "is_private", "users", "tickers")

    def __init__(self, name, is_private=False):

        self.name = name
        self.is_private = is_private
        self.users = set()
        self.tickers = {}


class PrivateRoom:
    __slots__ = ("name", "owner", "members", "operators")

    def __init__(self, name):

        self.name = name
        self.owner = None
        self.members = set()
        self.operators = set()


class ChatRooms:
    __slots__ = ("completions", "server_rooms", "joined_rooms", "private_rooms")

    # Trailing spaces to avoid conflict with regular rooms
    GLOBAL_ROOM_NAME = "Public "
    ROOM_NAME_MAX_LENGTH = 24

    def __init__(self):

        self.completions = set()
        self.server_rooms = set()
        self.joined_rooms = {}
        self.private_rooms = {}

        for event_name, callback in (
            ("global-room-message", self._global_room_message),
            ("join-room", self._join_room),
            ("leave-room", self._leave_room),
            ("private-room-add-operator", self._private_room_add_operator),
            ("private-room-add-user", self._private_room_add_user),
            ("private-room-added", self._private_room_added),
            ("private-room-operator-added", self._private_room_operator_added),
            ("private-room-operator-removed", self._private_room_operator_removed),
            ("private-room-operators", self._private_room_operators),
            ("private-room-remove-operator", self._private_room_remove_operator),
            ("private-room-remove-user", self._private_room_remove_user),
            ("private-room-removed", self._private_room_removed),
            ("private-room-toggle", self._private_room_toggle),
            ("private-room-users", self._private_room_users),
            ("quit", self._quit),
            ("room-list", self._room_list),
            ("say-chat-room", self._say_chat_room),
            ("server-login", self._server_login),
            ("server-disconnect", self._server_disconnect),
            ("start", self._start),
            ("ticker-add", self._ticker_add),
            ("ticker-remove", self._ticker_remove),
            ("ticker-state", self._ticker_state),
            ("user-joined-room", self._user_joined_room),
            ("user-left-room", self._user_left_room)
        ):
            events.connect(event_name, callback)

    def _start(self):

        for room in config.sections["server"]["autojoin"]:
            if isinstance(room, str):
                self.show_room(room, is_private=(room in self.private_rooms), switch_page=False, remembered=True)

    def _quit(self):
        self.remove_all_rooms(is_permanent=False)
        self.completions.clear()

    def _server_login(self, msg):

        if not msg.success:
            return

        # Request a complete room list. A limited room list not including blacklisted rooms and
        # rooms with few users is automatically sent when logging in, but subsequent room list
        # requests contain all rooms.
        self.request_room_list()

        self.request_private_room_toggle(config.sections["server"]["private_chatrooms"])

        for room in self.joined_rooms:
            if room == self.GLOBAL_ROOM_NAME:
                core.send_message_to_server(JoinGlobalRoom())
            else:
                core.send_message_to_server(JoinRoom(room))

    def _server_disconnect(self, _msg):

        for room_obj in self.joined_rooms.values():
            room_obj.tickers.clear()
            room_obj.users.clear()

        self.server_rooms.clear()
        self.private_rooms.clear()
        self.update_completions()

    def show_room(self, room, is_private=False, switch_page=True, remembered=False):

        room_obj = self.joined_rooms.get(room)

        if room_obj is None:
            self.joined_rooms[room] = room_obj = JoinedRoom(name=room, is_private=is_private)

            if room not in config.sections["server"]["autojoin"]:
                position = 0 if room == self.GLOBAL_ROOM_NAME else -1
                config.sections["server"]["autojoin"].insert(position, room)

        if not room_obj.users:
            if room == self.GLOBAL_ROOM_NAME:
                core.send_message_to_server(JoinGlobalRoom())
            else:
                core.send_message_to_server(JoinRoom(room, is_private))

        events.emit("show-room", room, is_private, switch_page, remembered)

    def remove_room(self, room, is_permanent=True):

        if room not in self.joined_rooms:
            return

        if room == self.GLOBAL_ROOM_NAME:
            core.send_message_to_server(LeaveGlobalRoom())
        else:
            core.send_message_to_server(LeaveRoom(room))

        room_obj = self.joined_rooms.pop(room)

        for username in room_obj.users:
            core.users.unwatch_user(username, context=f"chatrooms_{room}")

        if is_permanent:
            if room in config.sections["columns"]["chat_room"]:
                del config.sections["columns"]["chat_room"][room]

            if room in config.sections["server"]["autojoin"]:
                config.sections["server"]["autojoin"].remove(room)

        events.emit("remove-room", room)

    def remove_all_rooms(self, is_permanent=True):
        for room in self.joined_rooms.copy():
            self.remove_room(room, is_permanent)

    @classmethod
    def sanitize_room_name(cls, room):
        """Sanitize room name according to server requirements."""

        # Replace non-ASCII characters
        room = room.strip().encode("ascii", errors="replace").decode()

        # Remove two or more consecutive spaces
        room = " ".join(room.split())

        # Limit to 24 characters
        room = room[:cls.ROOM_NAME_MAX_LENGTH]
        return room

    def clear_room_messages(self, room):
        events.emit("clear-room-messages", room)

    def echo_message(self, room, message, message_type="local"):
        events.emit("echo-room-message", room, message, message_type)

    def send_message(self, room, message):

        if room not in self.joined_rooms:
            return

        event = core.pluginhandler.outgoing_public_chat_event(room, message)
        if event is None:
            return

        room, message = event

        if config.sections["words"]["replacewords"]:
            for word, replacement in config.sections["words"]["autoreplaced"].items():
                message = message.replace(str(word), str(replacement))

        core.send_message_to_server(SayChatroom(room, message))
        core.pluginhandler.outgoing_public_chat_notification(room, message)

    def add_user_to_private_room(self, room, username):
        core.send_message_to_server(PrivateRoomAddUser(room, username))

    def add_operator_to_private_room(self, room, username):
        core.send_message_to_server(PrivateRoomAddOperator(room, username))

    def remove_user_from_private_room(self, room, username):
        core.send_message_to_server(PrivateRoomRemoveUser(room, username))

    def remove_operator_from_private_room(self, room, username):
        core.send_message_to_server(PrivateRoomRemoveOperator(room, username))

    def is_private_room_owned(self, room):
        private_room = self.private_rooms.get(room)
        return private_room is not None and private_room.owner == core.users.login_username

    def is_private_room_member(self, room):
        return room in self.private_rooms

    def is_private_room_operator(self, room):
        private_room = self.private_rooms.get(room)
        return private_room is not None and core.users.login_username in private_room.operators

    def request_room_list(self):
        core.send_message_to_server(RoomList())

    def request_private_room_disown(self, room):

        if not self.is_private_room_owned(room):
            return

        core.send_message_to_server(PrivateRoomDisown(room))
        del self.private_rooms[room]

    def request_private_room_cancel_membership(self, room):

        if not self.is_private_room_member(room):
            return

        core.send_message_to_server(PrivateRoomCancelMembership(room))
        del self.private_rooms[room]

    def request_private_room_toggle(self, enabled):
        core.send_message_to_server(PrivateRoomToggle(enabled))

    def request_update_ticker(self, room, message):
        core.send_message_to_server(RoomTickerSet(room, message))

    def _update_room_user(self, room_obj, user_data):

        username = user_data.username
        room_obj.users.add(username)
        core.users.watch_user(username, context=f"chatrooms_{room_obj.name}", is_implicit=True)

        watched_user = core.users.watched[username]
        watched_user.upload_speed = user_data.avgspeed
        watched_user.files = user_data.files
        watched_user.folders = user_data.dirs

        core.users.statuses[username] = user_data.status

        # Request user's IP address, so we can get the country and ignore messages by IP
        if username not in core.users.addresses:
            core.users.request_ip_address(username)

        # Replace server-provided country with our own
        if username in core.users.countries:
            user_data.country = core.users.countries[username]

    def _update_private_room(self, room, owner=None, members=None, operators=None):

        private_room = self.private_rooms.get(room)

        if private_room is None:
            private_room = self.private_rooms[room] = PrivateRoom(room)

        if owner:
            private_room.owner = owner

        if members:
            for member in members:
                private_room.members.add(member)

        if operators:
            for operator in operators:
                private_room.operators.add(operator)

    def _join_room(self, msg):
        """Server code 14."""

        room_obj = self.joined_rooms.get(msg.room)

        if room_obj is None:
            # Reject unsolicited room join messages from the server
            core.send_message_to_server(LeaveRoom(msg.room))
            return

        room_obj.is_private = msg.private
        self.server_rooms.add(msg.room)

        if msg.private:
            self._update_private_room(msg.room, owner=msg.owner, operators=msg.operators)

        for user_data in msg.users:
            self._update_room_user(room_obj, user_data)

        core.pluginhandler.join_chatroom_notification(msg.room)

    def _leave_room(self, msg):
        """Server code 15."""

        room_obj = self.joined_rooms.get(msg.room)

        if room_obj is not None:
            for username in room_obj.users:
                core.users.unwatch_user(username, context=f"chatrooms_{msg.room}")

            room_obj.users.clear()

        core.pluginhandler.leave_chatroom_notification(msg.room)

    def _private_room_users(self, msg):
        """Server code 133."""

        self._update_private_room(msg.room, members=msg.users)

    def _private_room_add_user(self, msg):
        """Server code 134."""

        private_room = self.private_rooms.get(msg.room)

        if private_room is not None:
            private_room.members.add(msg.user)

    def _private_room_remove_user(self, msg):
        """Server code 135."""

        private_room = self.private_rooms.get(msg.room)

        if private_room is not None:
            private_room.members.discard(msg.user)

    def _private_room_added(self, msg):
        """Server code 139."""

        if msg.room in self.private_rooms:
            return

        self._update_private_room(msg.room)

        if msg.room in self.joined_rooms:
            # Room tab previously opened, join room now
            self.show_room(msg.room, is_private=True, switch_page=False)

        log.add(_("You have been added to a private room: %(room)s"), {"room": msg.room})

    def _private_room_removed(self, msg):
        """Server code 140."""

        if msg.room in self.private_rooms:
            del self.private_rooms[msg.room]

    def _private_room_toggle(self, msg):
        """Server code 141."""

        config.sections["server"]["private_chatrooms"] = msg.enabled

    def _private_room_add_operator(self, msg):
        """Server code 143."""

        private_room = self.private_rooms.get(msg.room)

        if private_room is not None:
            private_room.operators.add(msg.user)

    def _private_room_remove_operator(self, msg):
        """Server code 144."""

        private_room = self.private_rooms.get(msg.room)

        if private_room is not None:
            private_room.operators.discard(msg.user)

    def _private_room_operator_added(self, msg):
        """Server code 145."""

        private_room = self.private_rooms.get(msg.room)

        if private_room is not None:
            private_room.operators.add(core.users.login_username)

    def _private_room_operator_removed(self, msg):
        """Server code 146."""

        private_room = self.private_rooms.get(msg.room)

        if private_room is not None:
            private_room.operators.discard(core.users.login_username)

    def _private_room_operators(self, msg):
        """Server code 148."""

        self._update_private_room(msg.room, operators=msg.operators)

    def _global_room_message(self, msg):
        """Server code 152."""

        self._say_chat_room(msg, is_global=True)

    def _room_list(self, msg):
        """Server code 64."""

        for room, _user_count in msg.rooms:
            self.server_rooms.add(room)

        for room, _user_count in msg.ownedprivaterooms:
            self._update_private_room(room, owner=core.users.login_username)

        for room, _user_count in msg.otherprivaterooms:
            self._update_private_room(room)

        if config.sections["words"]["roomnames"]:
            self.update_completions()
            core.privatechat.update_completions()

    def get_message_type(self, user, text):

        if text.startswith("/me "):
            return "action"

        if user == core.users.login_username:
            return "local"

        if core.users.login_username and find_whole_word(core.users.login_username.lower(), text.lower()) > -1:
            return "hilite"

        return "remote"

    def _say_chat_room(self, msg, is_global=False):
        """Server code 13."""

        room = msg.room
        username = msg.user

        if not is_global:
            if room not in self.joined_rooms:
                msg.room = None
                return

            log.add_chat(_("Chat message from user '%(user)s' in room '%(room)s': %(message)s"), {
                "user": username,
                "room": room,
                "message": msg.message
            })

            if username != "server":
                if core.network_filter.is_user_ignored(username):
                    msg.room = None
                    return

                if core.network_filter.is_user_ip_ignored(username):
                    msg.room = None
                    return

            event = core.pluginhandler.incoming_public_chat_event(room, username, msg.message)
            if event is None:
                msg.room = None
                return

            _room, _username, msg.message = event
        else:
            room = self.GLOBAL_ROOM_NAME

        message = msg.message
        msg.message_type = self.get_message_type(username, message)
        is_action_message = (msg.message_type == "action")

        if is_action_message:
            message = message.replace("/me ", "", 1)

        if config.sections["words"]["censorwords"] and username != core.users.login_username:
            message = censor_text(message, censored_patterns=config.sections["words"]["censored"])

        if is_action_message:
            msg.formatted_message = msg.message = f"* {username} {message}"
        else:
            msg.formatted_message = f"[{username}] {message}"

        if is_global:
            msg.formatted_message = f"{msg.room} | {msg.formatted_message}"

        if config.sections["logging"]["chatrooms"] or room in config.sections["logging"]["rooms"]:
            log.write_log_file(
                folder_path=log.room_folder_path,
                basename=room, text=msg.formatted_message
            )

        if is_global:
            core.pluginhandler.public_room_message_notification(msg.room, username, msg.message)
        else:
            core.pluginhandler.incoming_public_chat_notification(room, username, msg.message)

    def _user_joined_room(self, msg):
        """Server code 16."""

        room_obj = self.joined_rooms.get(msg.room)

        if room_obj is None:
            msg.room = None
            return

        self._update_room_user(room_obj, msg.userdata)
        core.pluginhandler.user_join_chatroom_notification(msg.room, msg.userdata.username)

    def _user_left_room(self, msg):
        """Server code 17."""

        room_obj = self.joined_rooms.get(msg.room)

        if room_obj is None:
            msg.room = None
            return

        username = msg.username
        room_obj.users.discard(username)
        core.users.unwatch_user(username, context=f"chatrooms_{msg.room}")

        core.pluginhandler.user_leave_chatroom_notification(msg.room, username)

    def _ticker_state(self, msg):
        """Server code 113."""

        room_obj = self.joined_rooms.get(msg.room)

        if room_obj is None:
            msg.room = None
            return

        room_obj.tickers.clear()

        for username, message in msg.msgs:
            if core.network_filter.is_user_ignored(username) or \
                    core.network_filter.is_user_ip_ignored(username):
                # User ignored, ignore ticker message
                continue

            room_obj.tickers[username] = message

    def _ticker_add(self, msg):
        """Server code 114."""

        room_obj = self.joined_rooms.get(msg.room)

        if room_obj is None:
            msg.room = None
            return

        username = msg.user

        if core.network_filter.is_user_ignored(username) or core.network_filter.is_user_ip_ignored(username):
            # User ignored, ignore Ticker messages
            return

        room_obj.tickers[username] = msg.msg

    def _ticker_remove(self, msg):
        """Server code 115."""

        room_obj = self.joined_rooms.get(msg.room)

        if room_obj is None:
            msg.room = None
            return

        room_obj.tickers.pop(msg.user, None)

    def update_completions(self):

        self.completions.clear()
        self.completions.add(config.sections["server"]["login"])

        if config.sections["words"]["roomnames"]:
            self.completions.update(self.server_rooms)

        if config.sections["words"]["buddies"]:
            self.completions.update(core.buddies.users)

        if config.sections["words"]["commands"]:
            self.completions.update(core.pluginhandler.get_command_list("chatroom"))

        events.emit("room-completions", self.completions.copy())



# COPYRIGHT (C) 2020-2025 Nicotine+ Contributors
#
# GNU GENERAL PUBLIC LICENSE
#    Version 3, 29 June 2007
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys

from gi.repository import Gtk

import pynicotine
from pynicotine.core import core
from pynicotine.events import events
from pynicotine.gtkgui.application import GTK_API_VERSION
from pynicotine.gtkgui.widgets import ui
from pynicotine.gtkgui.widgets.dialogs import Dialog
from pynicotine.gtkgui.widgets.theme import add_css_class
from pynicotine.gtkgui.widgets.theme import remove_css_class
from pynicotine.utils import open_uri


class About(Dialog):

    AUTHORS = [
        "<b>Nicotine+ Team</b>",

        ("<b>Mat (mathiascode)</b>"
         "\n •  Maintainer (2020–present)"
         "\n •  Developer"),

        ("<b>Adam Cécile (eLvErDe)</b>"
         "\n •  Maintainer (2013–2016)"
         "\n •  Domain name administrator"
         "\n •  Source code migration from SVN to GitHub"
         "\n •  Developer"),

        ("<b>Han Boetes</b>"
         "\n •  Tester"
         "\n •  Documentation"
         "\n •  Bug hunting"
         "\n •  Translation management"),

        ("<b>alekksander</b>"
         "\n •  Tester"
         "\n •  Redesign of some graphics"),

        ("<b>slook</b>"
         "\n •  Tester"
         "\n •  Accessibility improvements"),

        ("<b>ketacat</b>"
         "\n •  Tester"),


        "\n<b>Nicotine+ Team (Emeritus)</b>",

        ("<b>daelstorm</b>"
         "\n •  Maintainer (2004–2009)"
         "\n •  Developer"),

        ("<b>quinox</b>"
         "\n •  Maintainer (2009–2012)"
         "\n •  Developer"),

        ("<b>Michael Labouebe (gfarmerfr)</b>"
         "\n •  Maintainer (2016–2017)"
         "\n •  Developer"),

        ("<b>Kip Warner</b>"
         "\n •  Maintainer (2018–2020)"
         "\n •  Developer"
         "\n •  Debianization"),

        ("<b>gallows (aka 'burp O')</b>"
         "\n •  Developer"
         "\n •  Packager"
         "\n •  Submitted Slack.Build file"),

        ("<b>hedonist (formerly known as alexbk)</b>"
         "\n •  OS X Nicotine.app maintainer / developer"
         "\n •  Author of PySoulSeek, used for Nicotine core"),

        ("<b>lee8oi</b>"
         "\n •  Bash commander"
         "\n •  New and updated /alias"),

        ("<b>INMCM</b>"
         "\n •  Nicotine+ topic maintainer on ubuntuforums.org"),

        ("<b>suser-guru</b>"
         "\n •  Suse Linux packager"
         "\n •  Nicotine+ RPM's for Suse 9.1, 9.2, 9.3, 10.0, 10.1"),

        ("<b>osiris</b>"
         "\n •  Handy-man"
         "\n •  Documentation"
         "\n •  Some GNU/Linux packaging"
         "\n •  Nicotine+ on Win32"
         "\n •  Author of Nicotine+ guide"),

        ("<b>Mutnick</b>"
         "\n •  Created Nicotine+ GitHub organization"
         "\n •  Developer"),

        ("<b>Lene Preuss</b>"
         "\n •  Python 3 migration"
         "\n •  Unit and DEP-8 continuous integration testing"),


        "\n<b>Nicotine Team (Emeritus)</b>",

        ("<b>Ingmar K. Steen (Hyriand)</b>"
         "\n •  Maintainer (2003–2004)"),

        ("<b>daelstorm</b>"
         "\n •  Beta tester"
         "\n •  Designer of most of the settings"
         "\n •  Made the Nicotine icons"),

        ("<b>SmackleFunky</b>"
         "\n •  Beta tester"),

        ("<b>Wretched</b>"
         "\n •  Beta tester"
         "\n •  Bringer of great ideas"),

        ("<b>(va)\\*10^3</b>"
         "\n •  Beta tester"
         "\n •  Designer of Nicotine homepage and artwork (logos)"),

        ("<b>sierracat</b>"
         "\n •  MacOSX tester"
         "\n •  soulseeX developer"),

        ("<b>Gustavo J. A. M. Carneiro</b>"
         "\n •  Created the exception dialog"),

        ("<b>SeeSchloss</b>"
         "\n •  Developer"
         "\n •  Created 1.0.8 Win32 installer"
         "\n •  Created Soulfind, open source Soulseek server written in D"),

        ("<b>vasi</b>"
         "\n •  Mac developer"
         "\n •  Packaged Nicotine on OSX PowerPC"),


        "\n<b>PySoulSeek Team (Emeritus)</b>",

        ("<b>Alexander Kanavin</b>"
         "\n •  Maintainer (2001–2003)"),

        ("<b>Nir Arbel</b>"
         "\n •  Helped with many protocol questions, and of course he designed and implemented the whole system"),

        ("<b>Brett W. Thompson (Zip)</b>"
         "\n •  His client code was used to get an initial impression of how the system works"
         "\n •  Supplied the patch for logging chat conversations"),

        ("<b>Josselin Mouette</b>"
         "\n •  Official Debian package maintainer"),

        ("<b>blueboy</b>"
         "\n •  Former unofficial Debian package maintainer"),

        ("<b>Christian Swinehart</b>"
         "\n •  Fink package maintainer"),

        ("<b>Ingmar K. Steen (Hyriand)</b>"
         "\n •  Patches for upload bandwidth management, banning, various UI improvements and more"),

        ("<b>Geert Kloosterman</b>"
         "\n •  A script for importing Windows Soulseek configuration"),

        ("<b>Joe Halliwell</b>"
         "\n •  Submitted a patch for optionally discarding search results after closing a search tab"),

        ("<b>Alexey Vyskubov</b>"
         "\n •  Code cleanups"),

        ("<b>Jason Green (SmackleFunky)</b>"
         "\n •  Ignore list and auto-join checkbox, wishlists")]

    TRANSLATORS = [
        ("<b>Albanian</b>"
         "\n •  W L (2023–2024)"),

        ("<b>Arabic</b>"
         "\n •  ButterflyOfFire (2024)"),

        ("<b>Catalan</b>"
         "\n •  Aniol (2024–2025)"
         "\n •  Maite Guix (2022)"),

        ("<b>Chinese (Simplified)</b>"
         "\n •  Ys413 (2024)"
         "\n •  Bonislaw (2023)"
         "\n •  hylau (2023)"
         "\n •  hadwin (2022)"),

        ("<b>Czech</b>"
         "\n •  slrslr (2024–2025)"
         "\n •  burnmail123 (2021–2023)"),

        ("<b>Danish</b>"
         "\n •  mathsped (2003–2004)"),

        ("<b>Dutch</b>"
         "\n •  Toine Rademacher (toineenzo) (2023–2024)"
         "\n •  Han Boetes (hboetes) (2021–2024)"
         "\n •  Kenny Verstraete (2009)"
         "\n •  nince78 (2007)"
         "\n •  Ingmar K. Steen (Hyriand) (2003–2004)"),

        ("<b>English</b>"
         "\n •  slook (2021–2024)"
         "\n •  Han Boetes (hboetes) (2021–2024)"
         "\n •  Mat (mathiascode) (2020–2024)"
         "\n •  Michael Labouebe (gfarmerfr) (2016)"
         "\n •  daelstorm (2004–2009)"
         "\n •  Ingmar K. Steen (Hyriand) (2003–2004)"),

        ("<b>Esperanto</b>"
         "\n •  phlostically (2021)"),

        ("<b>Estonian</b>"
         "\n •  rimasx (2024)"
         "\n •  PriitUring (2023)"),

        ("<b>Euskara</b>"
         "\n •  Julen (2006–2007)"),

        ("<b>Finnish</b>"
         "\n •  Kari Viittanen (Kalevi) (2006–2007)"),

        ("<b>French</b>"
         "\n •  Saumon (2023)"
         "\n •  subu_versus (2023)"
         "\n •  zniavre (2007–2023)"
         "\n •  Maxime Leroy (Lisapple) (2021–2022)"
         "\n •  Mohamed El Morabity (melmorabity) (2021–2024)"
         "\n •  m-balthazar (2020)"
         "\n •  Michael Labouebe (gfarmerfr) (2016–2017)"
         "\n •  Monsieur Poisson (2009–2010)"
         "\n •  ManWell (2007)"
         "\n •  systr (2006)"
         "\n •  Julien Wajsberg (flashfr) (2003–2004)"),

        ("<b>German</b>"
         "\n •  Han Boetes (hboetes) (2021–2024)"
         "\n •  phelissimo_ (2023)"
         "\n •  Meokater (2007)"
         "\n •  (._.) (2007)"
         "\n •  lippel (2004)"
         "\n •  Ingmar K. Steen (Hyriand) (2003–2004)"),

        ("<b>Hungarian</b>"
         "\n •  Szia Tomi (2022–2024)"
         "\n •  Nils (2009)"
         "\n •  David Balazs (djbaloo) (2006–2020)"),

        ("<b>Italian</b>"
         "\n •  Gabriele (Gabboxl) (2022–2023)"
         "\n •  ms-afk (2023)"
         "\n •  Gianluca Boiano (2020–2023)"
         "\n •  nicola (2007)"
         "\n •  dbazza (2003–2004)"),

        ("<b>Latvian</b>"
         "\n •  Pagal3 (2022–2025)"),

        ("<b>Lithuanian</b>"
         "\n •  mantas (2020)"
         "\n •  Žygimantas Beručka (2006–2009)"),

        ("<b>Norwegian Bokmål</b>"
         "\n •  Allan Nordhøy (comradekingu) (2021)"),

        ("<b>Polish</b>"
         "\n •  Mariusz (mariachini) (2017–2024)"
         "\n •  Amun-Ra (2007)"
         "\n •  thine (2007)"
         "\n •  Wojciech Owczarek (owczi) (2003–2004)"),

        ("<b>Portuguese (Brazil)</b>"
         "\n •  Havokdan (2022–2023)"
         "\n •  Guilherme Santos (2022)"
         "\n •  b1llso (2022)"
         "\n •  Nicolas Abril (2021)"
         "\n •  yyyyyyyan (2020)"
         "\n •  Felipe Nogaroto Gonzalez (Suicide|Solution) (2006)"),

        ("<b>Portuguese (Portugal)</b>"
         "\n •  ssantos (2023)"
         "\n •  Vinícius Soares (2023)"),

        ("<b>Romanian</b>"
         "\n •  Slendi (xslendix) (2023)"),

        ("<b>Russian</b>"
         "\n •  Kirill Feoktistov (SnIPeRSnIPeR) (2022–2024)"
         "\n •  Mehavoid (2021–2023)"
         "\n •  AHOHNMYC (2022)"),

        ("<b>Slovak</b>"
         "\n •  Jozef Říha (2006–2008)"),

        ("<b>Spanish (Chile)</b>"
         "\n •  MELERIX (2021–2023)"
         "\n •  tagomago (2021–2022)"
         "\n •  Strange (2021)"
         "\n •  Silvio Orta (2007)"
         "\n •  Dreslo (2003–2004)"),

        ("<b>Spanish (Spain)</b>"
         "\n •  gallegonovato (2023–2024)"
         "\n •  MELERIX (2021–2023)"
         "\n •  tagomago (2021–2022)"
         "\n •  Strange (2021)"
         "\n •  Silvio Orta (2007)"
         "\n •  Dreslo (2003–2004)"),

        ("<b>Swedish</b>"
         "\n •  mitramai (2021)"
         "\n •  Markus Magnuson (alimony) (2003–2004)"),

        ("<b>Tamil</b>"
         "\n •  தமிழ்நேரம் (2025)"),

        ("<b>Turkish</b>"
         "\n •  Oğuz Ersen (2021–2024)"),

        ("<b>Ukrainian</b>"
         "\n •  Oleg Gritsun (2024–2025)"
         "\n •  uniss2209 (2022)")]

    LICENSE = [
        ("Nicotine+ is licensed under the <a href='https://www.gnu.org/licenses/gpl-3.0.html'>"
         "GNU General Public License v3.0 or later</a>, with the following exceptions:"),

        ("<b><a href='https://github.com/tinytag/tinytag'>tinytag</a> licensed under the MIT License.</b>"
         "\nCopyright (c) 2014-2023 Tom Wallroth, Mat (mathiascode)"
         "\n\nPermission is hereby granted, free of charge, to any person obtaining a copy "
         'of this software and associated documentation files (the "Software"), to deal '
         "in the Software without restriction, including without limitation the rights "
         "to use, copy, modify, merge, publish, distribute, sublicense, and/or sell "
         "copies of the Software, and to permit persons to whom the Software is "
         "furnished to do so, subject to the following conditions: "
         "\n\nThe above copyright notice and this permission notice shall be included in all "
         "copies or substantial portions of the Software."),

        ("<b><a href='https://github.com/madebybowtie/FlagKit'>FlagKit</a> icons licensed "
         "under the MIT License.</b>"
         "\nCopyright (c) 2016 Bowtie AB"
         "\n\nPermission is hereby granted, free of charge, to any person obtaining a copy "
         'of this software and associated documentation files (the "Software"), to deal '
         "in the Software without restriction, including without limitation the rights "
         "to use, copy, modify, merge, publish, distribute, sublicense, and/or sell "
         "copies of the Software, and to permit persons to whom the Software is "
         "furnished to do so, subject to the following conditions: "
         "\n\nThe above copyright notice and this permission notice shall be included in all "
         "copies or substantial portions of the Software."),

        ("<b>IP2Location country data licensed under the "
         "<a href='https://creativecommons.org/licenses/by-sa/4.0/'>CC-BY-SA-4.0 License</a>.</b>"
         "\nCopyright (c) 2001–2024 Hexasoft Development Sdn. Bhd."
         "\nNicotine+ uses the IP2Location LITE database for "
         "<a href='https://lite.ip2location.com'>IP geolocation</a>.")]

    def __init__(self, application):

        (
            self.application_version_label,
            self.authors_container,
            self.container,
            self.copyright_label,
            self.dependency_versions_label,
            self.license_container,
            self.main_icon,
            self.status_container,
            self.status_icon,
            self.status_label,
            self.status_spinner,
            self.translators_container,
            self.website_label
        ) = ui.load(scope=self, path="dialogs/about.ui")

        super().__init__(
            parent=application.window,
            content_box=self.container,
            show_callback=self.on_show,
            close_callback=self.on_close,
            title=_("About"),
            width=425,
            height=540,
            show_title=False
        )

        self.is_version_outdated = False

        icon_name = pynicotine.__application_id__
        icon_args = (Gtk.IconSize.BUTTON,) if GTK_API_VERSION == 3 else ()  # pylint: disable=no-member
        website_text = _('Website')
        gtk_version = f"{Gtk.get_major_version()}.{Gtk.get_minor_version()}.{Gtk.get_micro_version()}"

        self.main_icon.set_from_icon_name(icon_name, *icon_args)
        self.website_label.connect("activate-link", self.on_activate_link)

        for label_widget, text in (
            (self.application_version_label, f"{pynicotine.__application_name__} {pynicotine.__version__}"),
            (self.dependency_versions_label, (f"GTK {gtk_version}   •   Python {sys.version.split()[0]}")),
            (self.website_label, (f"<a href='{pynicotine.__website_url__}' title='{pynicotine.__website_url__}'>"
                                  f"{website_text}</a>")),
            (self.copyright_label, f"<small>{pynicotine.__copyright__}</small>")
        ):
            label_widget.set_markup(text)

        for entries, container in (
            (self.AUTHORS, self.authors_container),
            (self.TRANSLATORS, self.translators_container),
            (self.LICENSE, self.license_container)
        ):
            for text in entries:
                label = Gtk.Label(label=text, use_markup=True, selectable=True, wrap=True, xalign=0, visible=True)

                if entries is self.LICENSE:
                    label.connect("activate-link", self.on_activate_link)

                if GTK_API_VERSION >= 4:
                    container.append(label)  # pylint: disable=no-member
                else:
                    container.add(label)     # pylint: disable=no-member

        events.connect("check-latest-version", self.on_check_latest_version)

    def on_activate_link(self, _label, url):
        open_uri(url)
        return True

    def on_check_latest_version(self, latest_version, is_outdated, error):

        if error:
            icon_name = "emblem-important-symbolic"
            css_class = "error"
            message = _("Error checking latest version: %s") % error

        elif is_outdated:
            icon_name = "dialog-warning-symbolic"
            css_class = "warning"
            message = _("New release available: %s") % latest_version

        else:
            icon_name = "object-select-symbolic"
            css_class = "success"
            message = _("Up to date")

        self.is_version_outdated = is_outdated
        icon_args = (Gtk.IconSize.BUTTON,) if GTK_API_VERSION == 3 else ()  # pylint: disable=no-member

        self.status_label.set_label(message)

        self.status_icon.set_from_icon_name(icon_name, *icon_args)
        self.status_spinner.set_visible(False)
        self.status_spinner.stop()

        add_css_class(self.status_container, css_class)

    def on_show(self, *_args):

        if core.update_checker is None:
            # Update checker is not loaded
            return

        if self.is_version_outdated:
            # No need to check latest version again
            return

        if not self.is_visible():
            return

        for css_class in ("error", "warning", "success"):
            remove_css_class(self.status_container, css_class)

        self.status_label.set_label(_("Checking latest version…"))
        self.status_spinner.set_visible(True)
        self.status_spinner.start()
        self.status_container.set_visible(True)

        core.update_checker.check()

    def on_close(self, *_args):

        self.main_icon.grab_focus()
        self.status_spinner.stop()
        self.container.get_vadjustment().set_value(0)
